# Copyright (C) 2017-2023 Pier Carlo Chiodi
# Copyright (C) 2021 Vilhelm Prytz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import ARouteServerCommand
from ..config.program import program_config

from ..config.general import ConfigParserGeneral
from ..config.clients import ConfigParserClients


class CheckConfigCommand(ARouteServerCommand):
    COMMAND_NAME = "check-config"
    COMMAND_HELP = (
        "This command will check if the configuration "
        "files for ARouteServer are valid."
    )

    NEEDS_CONFIG = True

    @classmethod
    def add_arguments(cls, parser):
        super(CheckConfigCommand, cls).add_arguments(parser)

        group = parser.add_argument_group(
            title="Route server configuration",
            description="The following arguments override those provided "
            "in the program's configuration file.",
        )

        group.add_argument(
            "--general",
            help="General route server configuration file.",
            metavar="FILE",
            dest="cfg_general",
        )

        group.add_argument(
            "--clients",
            help="Route server clients configuration file.",
            metavar="FILE",
            dest="cfg_clients",
        )

    def check_config(self):
        general = ConfigParserGeneral()
        general.load(self.general_path)
        general.parse()

        clients = ConfigParserClients(general_cfg=general)
        clients.load(self.clients_path)
        clients.parse()

    def run(self):
        self.general_path = program_config.get("cfg_general")
        self.clients_path = program_config.get("cfg_clients")

        if self.args.cfg_general:
            self.general_path = self.args.cfg_general

        if self.args.cfg_clients:
            self.clients_path = self.args.cfg_clients

        logging.info(f"Checking configuration files {self.general_path} and {self.clients_path}...")

        self.check_config()

        return True



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import json
import re
import os
import threading

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError, RetryError
from urllib3.util.retry import Retry

from .cached_objects import CachedObject
from .config.validators import ValidatorASSet
from .errors import PeeringDBError, PeeringDBNoInfoError, ConfigError
from .irrdb import IRRDBInfo
from .version import __version__


session_cache = {}


class TimeoutHTTPAdapter(HTTPAdapter):

    def __init__(self, *args, **kwargs) :
        self.timeout = 30

        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]

        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")

        if timeout is None:
            kwargs["timeout"] = self.timeout

        return super().send(request, **kwargs)


def _get_request_session():
    thread_id = threading.get_ident()

    if thread_id in session_cache:
        return session_cache[thread_id]

    retry_strategy = Retry(
        total=3,
        backoff_factor=2,
        status_forcelist=[413, 429, 500, 502, 503, 504],
        allowed_methods=["HEAD", "GET", "POST", "OPTIONS"],

        # This is to avoid we wait for minutes in case of a 429 with a long
        # Retry-After header.
        respect_retry_after_header=False
    )

    adapter = TimeoutHTTPAdapter(max_retries=retry_strategy)

    session = requests.Session()
    session.mount("https://", adapter)
    session.mount("http://", adapter)

    session_cache[thread_id] = session
    return session


class PeeringDBInfo(CachedObject):

    EXPIRY_TIME_TAG = "pdb_info"

    MISSING_INFO_EXCEPTION = PeeringDBNoInfoError

    PEERING_DB_API_KEY_ENV_VAR = "SECRET_PEERINGDB_API_KEY"

    PEERING_DB_API_KEY_WELL_KNOWN_FILES = (
        "~/.arouteserver/peeringdb_api.key",
        "~/.peeringdb_api.key"
    )

    def _get_peeringdb_url(self):
        raise NotImplementedError()

    @staticmethod
    def _read_from_url(url):
        headers = {
            "User-Agent": "arouteserver/{}".format(__version__)
        }

        peeringdb_api_key = None

        for env_var in (PeeringDBInfo.PEERING_DB_API_KEY_ENV_VAR, "PEERINGDB_API_KEY", "API_KEY"):
            if env_var in os.environ:
                peeringdb_api_key = os.environ[env_var].strip()
                break

        if not peeringdb_api_key:
            for well_known_file in PeeringDBInfo.PEERING_DB_API_KEY_WELL_KNOWN_FILES:
                path = os.path.expanduser(well_known_file)
                if os.path.exists(path):
                    with open(path, "r") as f:
                        peeringdb_api_key = f.read().strip()
                        break

        if peeringdb_api_key:
            headers["Authorization"] = "Api-Key {}".format(peeringdb_api_key)

        session = _get_request_session()

        try:
            response = session.get(url, headers=headers)
            response.raise_for_status()

        except (HTTPError, RetryError) as e:
            if isinstance(e, HTTPError) and e.response.status_code == 404:
                return "{}"
            else:
                additional_info = ""
                if (
                    (isinstance(e, HTTPError) and e.response.status_code == 429) or \
                    isinstance(e, RetryError)
                ):
                    additional_info = (
                        " - Please consider using a PeeringDB API key to perform "
                        "authentication, which could help mitigating the effects "
                        "of anonymous API query rate-limit. "
                        "The key can be configured by setting the environment "
                        "variable {} or can be stored inside one of the following "
                        "well-known files: {} "
                        "Documentation on how to create an API key can be found "
                        "on the peeringdb.com web site "
                        "(https://docs.peeringdb.com/howto/api_keys/)".format(
                            PeeringDBInfo.PEERING_DB_API_KEY_ENV_VAR,
                            ", ".join(PeeringDBInfo.PEERING_DB_API_KEY_WELL_KNOWN_FILES)
                        )
                    )

                raise PeeringDBError(
                    "HTTP error while retrieving info from PeeringDB: "
                    "{}{}".format(
                        str(e), additional_info
                    )
                )
        except Exception as e:
            raise PeeringDBError(
                "Error while retrieving info from PeeringDB: {}".format(
                    str(e)
                )
            )

        return response.content.decode("utf-8")

    def _get_data_from_peeringdb(self):
        plain_text = self._read_from_url(self._get_peeringdb_url())
        try:
            data = json.loads(plain_text)
            return data
        except Exception as e:
            raise PeeringDBError(
                "Error while decoding PeeringDB output: {}".format(
                    str(e)
                )
            )

    def _get_data_from_peeringdb_bulk_query(self):
        return None

    def _get_data(self):
        data = self._get_data_from_peeringdb_bulk_query()
        if data:
            return [data]

        data = self._get_data_from_peeringdb()
        if "data" not in data:
            raise PeeringDBNoInfoError("Missing 'data'")
        if not isinstance(data["data"], list):
            raise PeeringDBNoInfoError("Unexpected format: 'data' is not a list")
        if len(data["data"]) == 0:
            raise PeeringDBNoInfoError("No data for this network")

        return data["data"]

class PeeringDBNet(PeeringDBInfo):

    EXPIRY_TIME_TAG = "pdb_info"

    PEERINGDB_URL = "https://www.peeringdb.com/api/net?asn={asn}"
    PEERINGDB_BULK_QUERY_URL = "https://www.peeringdb.com/api/net?asn__in={asns}"

    BULK_QUERY_CACHE = {}

    def __init__(self, asn, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)
        self.asn = asn

    @classmethod
    def populate_bulk_query_cache(cls, asns):
        logging.debug("Pre-populating PeeringDB 'net' cache for {} ASNs...".format(len(asns)))

        def chunks(lst, n):
            """Yield successive n-sized chunks from lst."""
            for i in range(0, len(lst), n):
                yield lst[i:i + n]

        chunk_size = int(os.getenv("PEERINGDB_BULK_QUERY_CHUNK_SIZE", 150))

        for asns_group in chunks(asns, chunk_size):

            plain_text = cls._read_from_url(
                cls.PEERINGDB_BULK_QUERY_URL.format(
                    asns=",".join(map(str, asns_group))
                )
            )

            try:
                data = json.loads(plain_text)
            except Exception as e:
                raise PeeringDBError(
                    "Error while decoding PeeringDB output for bulk query: {}".format(
                        str(e)
                    )
                )

            for asn in asns_group:
                if asn not in cls.BULK_QUERY_CACHE:
                    cls.BULK_QUERY_CACHE[asn] = None

                for peering_db_net in data["data"]:
                    peering_db_net_asn = peering_db_net["asn"]
                    cls.BULK_QUERY_CACHE[peering_db_net_asn] = peering_db_net

    def _get_data_from_peeringdb_bulk_query(self):
        if self.asn in self.BULK_QUERY_CACHE:
            data = self.BULK_QUERY_CACHE[self.asn]
            if data:
                logging.debug("Bulk query cache hit for {}".format(self.asn))
                return data
            else:
                raise PeeringDBNoInfoError("No data for this network")

    def load_data(self):
        logging.debug("Getting data from PeeringDB: net {}".format(self.asn))

        PeeringDBInfo.load_data(self)

        self.info_prefixes4 = self.raw_data[0].get("info_prefixes4", None)
        self.info_prefixes6 = self.raw_data[0].get("info_prefixes6", None)
        self.irr_as_sets = self.parse_as_sets(
            self.raw_data[0].get("irr_as_set", None)
        )

    def parse_as_sets(self, raw_irr_as_sets):
        res = []
        if raw_irr_as_sets and raw_irr_as_sets.strip():
            raw_irr_as_sets = re.split(r"[/,&\s]", raw_irr_as_sets)
            for raw_irr_as_set in raw_irr_as_sets:
                irr_as_set = self.parse_as_set(raw_irr_as_set)
                if irr_as_set and irr_as_set not in res:
                    res.append(irr_as_set)
        return res

    def parse_as_set(self, in_value):
        v = in_value.strip()

        if not v:
            return None

        guessed = False

        # Removing "ipv4:" and "ipv6:".
        pattern = re.compile("^(?:ipv4|ipv6):", flags=re.IGNORECASE)
        v, number_of_subs_made = pattern.subn("", v)
        if number_of_subs_made > 0:
            v = v.strip()
            guessed = True
        if not v:
            return None

        # IRR record source
        valid_sources_regex = IRRDBInfo.BGPQ3_DEFAULT_SOURCES.replace(",", "|")

        # Converting stuff like AS-FOO@SOURCE in SOURCE::AS-FOO
        pattern = re.compile(
            "^([^@]+)@({sources})$".format(sources=valid_sources_regex),
            flags=re.IGNORECASE
        )
        v, number_of_subs_made = pattern.subn("\\2::\\1", v)
        if number_of_subs_made > 0:
            guessed = True

        # Converting "SOURCE:AS-FOO" format (single colon) to "SOURCE::AS-FOO"
        # (only for known sources)
        pattern = re.compile(
            "^({sources}):([^:].+)$".format(sources=valid_sources_regex),
            flags=re.IGNORECASE
        )
        v, number_of_subs_made = pattern.subn("\\1::\\2", v)
        if number_of_subs_made > 0:
            guessed = True

        try:
            v = ValidatorASSet().validate(v)
        except ConfigError as e:
            logging.debug("AS-SET from PeeringDB for AS{}: "
                          "ignoring {}, {}".format(self.asn, v, str(e)))
            return None

        if guessed:
            logging.info("AS-SET from PeeringDB for AS{}: "
                         "guessed {} from {}".format(self.asn, v, in_value))

        return v

    def _get_object_filename(self):
        return "peeringdb_net_{}.json".format(self.asn)

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL.format(asn=self.asn)

class PeeringDBNetIXLan(PeeringDBInfo):

    PEERINGDB_URL = "https://www.peeringdb.com/api/netixlan?ixlan_id={ixlanid}"

    def __init__(self, ixlanid, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)
        self.ixlanid = ixlanid

    def load_data(self):
        logging.debug("Getting data from PeeringDB: Net IX LAN {}".format(self.ixlanid))

        PeeringDBInfo.load_data(self)

    def _get_object_filename(self):
        return "peeringdb_ixlanid_{}.json".format(self.ixlanid)

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL.format(ixlanid=self.ixlanid)

class PeeringDBNetNeverViaRouteServers(PeeringDBInfo):

    PEERINGDB_URL = "https://www.peeringdb.com/api/net?info_never_via_route_servers=1"

    EXPIRY_TIME_TAG = "pdb_info"

    def __init__(self, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)

        self.networks = []

    def load_data(self):
        logging.debug("Getting 'never via route-servers' networks from PeeringDB")

        PeeringDBInfo.load_data(self)

        for network in self.raw_data:
            self.networks.append({
                "asn": network["asn"]
            })

    def _get_object_filename(self):
        return "peeringdb_neverviarouteservers.json"

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL.format()

class PeeringDBIXList(PeeringDBInfo):

    PEERINGDB_URL = "https://peeringdb.com/api/ix"

    def __init__(self, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)

        self.ixp_list = []

    def load_data(self):
        logging.debug("Getting the list of IXs from PeeringDB")

        PeeringDBInfo.load_data(self)

        for ixp in self.raw_data:
            self.ixp_list.append({
                "city": ixp["city"],
                "country": ixp["country"],
                "full_name": ixp["name_long"],
                "short_name": ixp["name"],
                "peeringdb_handle": ixp["id"]
            })

    def _get_object_filename(self):
        return "peeringdb_ixlist.json"

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL

def clients_from_peeringdb(netixlanid, cache_dir):
    clients = []

    pdb_net_ixlan = PeeringDBNetIXLan(netixlanid, cache_dir=cache_dir)
    pdb_net_ixlan.load_data()
    netixlans = pdb_net_ixlan.raw_data
    for netixlan in netixlans:
        if netixlan["is_rs_peer"] is True:
            client = {
                "asn": netixlan["asn"],
                "ip": [],
            }
            for ipver in ("ipaddr4", "ipaddr6"):
                if netixlan[ipver]:
                    client["ip"].append(netixlan[ipver].encode("ascii", "ignore").decode("utf-8"))
            clients.append(client)

    asns = {}

    for client in clients:
        asn = client["asn"]
        net = PeeringDBNet(asn, cache_dir=cache_dir)
        net.load_data()

        if not net.irr_as_sets:
            continue

        key = "AS{}".format(asn)
        if key not in asns:
            asns[key] = {
                "as_sets": []
            }

        for irr_as_set in net.irr_as_sets:
            irr_as_set = irr_as_set.strip()
            if irr_as_set not in asns[key]["as_sets"]:
                asns[key]["as_sets"].append(irr_as_set.encode("ascii", "ignore").decode("utf-8"))

    data = {
        "asns": asns,
        "clients": clients
    }

    return data



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from jinja2 import Environment, FileSystemLoader
import os
import re
import time
import yaml

from .docker import InstanceError

from pierky.arouteserver.ipaddresses import IPAddress, IPNetwork
from pierky.arouteserver.tests.base import ARouteServerTestCase
from pierky.arouteserver.tests.mocked_env import MockedEnv
from pierky.arouteserver.tests.live_tests.instances import BGPSpeakerInstance


class LiveScenario(ARouteServerTestCase):
    """An helper class to run tests for a given scenario.

    This class must be derived by scenario-specific classes that
    must:

    - set the ``MODULE_PATH`` attribute to ``__file__``, in order
      to correctly locate files needed by the scenario.

    - fill the ``INSTANCES`` list with a list of
      :class:`BGPSpeakerInstance`
      (or derived) instances representing all the BGP speakers involved in the
      scenario.

    - set the ``SHORT_DESCR`` with a short description of the
      scenario; it will be used to print tests description in the
      format "<SHORT_DESCR>: <test function docstring>".

    - fill the ``DATA`` dictionary with IP prefixes IDs that
      represent all the IP addresses used by the scenario; this
      is useful to decouple real IP addresses from what they
      represent or are used to and to have a single definition
      of the tests functions and scenario configuration that can
      be used for both IPv4 and IPv6 tests.

      Example::

        DATA = {
            "rs_IPAddress": "192.0.2.2",
            "AS1_IPAddress": "192.0.2.11",
            "AS1_allowed_prefixes": "1.0.0.0/8",
            "AS1_good_prefix": "1.0.1.0/24",
            "AS101_prefixes": "101.0.0.0/8"
        }

    - optionally, if it's needed by the scenario, the derived classes
      must also fill the ``AS_SET`` and ``R_SET`` dictionaries with
      the expected content of any expanded AS-SETs used in IRRDB
      validation:

      - ``AS_SET``'s items must be in the format
        ``<AS_SET_name>: <list_of_authorized_origin_ASNs>``.

      - ``R_SET``'s items must be in the format
        ``<AS_SET_name>: <list_of_authorized_prefix_IDs>`` (where prefix
        IDs are those reported in the ``DATA`` dictionary).

      Example::

        AS_SET = {
            "AS-AS1": [1],
            "AS-AS1_CUSTOMERS": [101],
            "AS-AS2": [2],
            "AS-AS2_CUSTOMERS": [101]
        }
        R_SET = {
            "AS-AS1": [
                "AS1_allowed_prefixes"
            ],
            "AS-AS1_CUSTOMERS": [
                "AS101_prefixes"
            ]
        }

    - optionally, if it's needed by the scenario, the derived classes
      can also set the ``RTT`` dictionary with the values of the RTTs
      of the clients. Keys can be IP addresses of clients or the
      prefix IDs used in the ``DATA`` dictionary to represent them.

      Example::

        RTT = {
            "192.0.2.11": 14,
            "2001:db8::11": 165,
            "AS1_IPAddress": 44
        }

    - optionally, if it's needed by the scenario, the derived classes
      can also set the ``REJECT_CAUSE_COMMUNITY`` attribute with the
      pattern followed by BGP communities used to tag routes that are
      considered to be rejected (see the ``filtering.reject_policy``
      general configuration section and the ``reject_cause``
      community).
      The value of this attribute must be a regular expression that
      matches the standard, extended or large BGP communities used to
      tag invalid routes. For example, if the standard BGP community
      ``65520:dyn_val`` is used, the value must be ``^65520:(\\d+)$``.
      If this attribute is not None, routes that have LOCAL_PREF == 1
      and the ``reject_cause`` BGP community with ``dyn_val == 0``
      are considered as filtered.
      The ``REJECTED_ROUTE_ANNOUNCED_BY_COMMUNITY`` attribute can also
      be set to match the community used to track the announcing ASN of
      invalid routes.

    - implement the ``set_instance_variables`` method, used to set
      local instance attributes for the instances used within the
      tests functions.

      Example::

        def set_instance_variables(self):
            self.AS1 = self._get_instance_by_name("AS1")
            self.AS2 = self._get_instance_by_name("AS2")
            self.AS131073 = self._get_instance_by_name("AS131073")
            self.rs = self._get_instance_by_name("rs")

    Set the ``__test__`` attribute of derived classes to False to avoid them
    to be used directly by ``pytest`` to run tests; only the specific IPv4/IPv6
    class (the one where the ``DATA`` dictionary is set) must have
    ``__test__`` == True.
    """

    MODULE_PATH = None
    SHORT_DESCR = ""

    DATA = {}
    AS_SET = {}
    R_SET = {}
    RTT = {}
    INSTANCES = []

    DEBUG = False
    DO_NOT_STOP_INSTANCES = False

    CONFIG_BUILDER_CLASS = None

    MOCK_PEERING_DB = True
    MOCK_RIPE_RPKI_CACHE = True
    MOCK_IRRDB = True
    MOCK_RTTGETTER = True
    MOCK_ARIN_DB_DUMP = True

    # regex: for example ^65520:(\d+)$
    REJECT_CAUSE_COMMUNITY = None
    REJECTED_ROUTE_ANNOUNCED_BY_COMMUNITY = None

    ALLOWED_LOG_ERRORS = []

    @classmethod
    def _get_module_dir(cls):
        return os.path.dirname(cls.MODULE_PATH)

    @classmethod
    def _get_local_file_path(cls, filename):
        return os.path.join(cls._get_module_dir(), filename)

    @classmethod
    def _get_instance_by_name(cls, name):
        for instance in cls.INSTANCES:
            if instance.name == name:
                return instance
        raise Exception("Instance not found: {}.".format(name))

    @classmethod
    def _create_var_dir(cls):
        var_dir = "{}/var".format(cls._get_module_dir())
        if not os.path.exists(var_dir):
            os.mkdir(var_dir)
        return var_dir

    @classmethod
    def _do_not_stop_instances(cls):
        return cls.DO_NOT_STOP_INSTANCES or "REUSE_INSTANCES" in os.environ

    @classmethod
    def build_other_cfg(cls, tpl_name):
        """Builds configuration files for BGP speakers which are not the route server.

        Args:
            tpl_name (str): the name of the Jinja2 template file
                relative to the scenario directory.

        Returns:
            the path of the local rendered file.

        To render the template, one attribute is used and consumed by Jinja2:

        - ``data``, the scenario's ``DATA`` dictionary.

        The resulting file is saved into the local ``var`` directory
        and its absolute path is returned.
        """
        cls.debug("Building config from {}/{}".format(
            cls._get_module_dir(), tpl_name))

        env = Environment(
            loader=FileSystemLoader(cls._get_module_dir()),
            trim_blocks=True,
            lstrip_blocks=True
        )
        tpl = env.get_template(tpl_name)
        cfg = tpl.render(data=cls.DATA)

        var_dir = cls._create_var_dir()
        cfg_file_path = "{}/{}.config".format(var_dir, tpl_name)

        with open(cfg_file_path, "w") as f:
            f.write(cfg)

        return cfg_file_path

    @classmethod
    def _get_cfg_general(cls, filename=None):
        return filename or "general.yml"

    @classmethod
    def build_rs_cfg(cls, tpl_dir_name, tpl_name, out_file_name, ip_ver,
                      cfg_general=None, cfg_bogons="bogons.yml",
                      cfg_clients="clients.yml", **kwargs):
        """Builds configuration file for the route server.

        Args:
            tpl_dir_name (str): the directory where Jinja2
                templates are located, relative to the current scenario.

            tpl_name (str): the name of the template to be
                rendered.

            out_file_name (str): the name of the destination
                file.

            ip_ver (int): the IP version for which this route server
                will operate. Use None if the configuration is valid
                for both IPv4 and IPv6.

            cfg_general (str), cfg_bogons (str), cfg_clients (str): the
                name of the 3 main files containing route server's
                options and policies, clients definition and bogons
                IP addresses. File names are relative to the scenario
                directory.

        Returns:
            the path of the local rendered file.

        The resulting file is saved into the local ``var`` directory
        and its absolute path is returned.
        """

        cls.debug("Building config from {}/{}".format(
            tpl_dir_name, tpl_name)
        )

        var_dir = cls._create_var_dir()

        builder = cls.CONFIG_BUILDER_CLASS(
            template_dir="{}/{}".format(cls._get_module_dir(), tpl_dir_name),
            template_name=tpl_name,
            cache_dir=var_dir,
            cfg_general="{}/{}".format(cls._get_module_dir(),
                                       cfg_general or cls._get_cfg_general()),
            cfg_bogons="{}/{}".format(cls._get_module_dir(), cfg_bogons),
            cfg_clients="{}/{}".format(cls._get_module_dir(), cfg_clients),
            ip_ver=ip_ver,
            ignore_errors=["*"],
            live_tests=True,
            **kwargs
        )

        cfg_file_path = "{}/{}".format(var_dir, out_file_name)

        with open(cfg_file_path, "w") as f:
            builder.render_template(f)

        cls.rs_cfg_file_path = cfg_file_path

        return cfg_file_path

    @classmethod
    def use_static_file(cls, local_filename):
        """Prepare the local file in order to use it later.

        Args:
            filename (str): the name of the local file,
                relative to the scenario directory.

        Returns:
            the path of the file to be used
        """

        var_dir = cls._create_var_dir()
        var_path = os.path.join(var_dir, local_filename)
        local_path = os.path.join(cls._get_module_dir(), local_filename)
        with open(local_path, "r") as src:
            with open(var_path, "w") as dst:
                dst.write(src.read())
        return var_path

    @classmethod
    def _setUpClass(cls):
        for prefix_name in cls.DATA:
            prefix = cls.DATA[prefix_name]
            net = IPNetwork(prefix) if "/" in prefix else IPAddress(prefix)
            if str(net) != prefix:
                raise ValueError(
                    "Prefix '{}' is not represented in its canonical form: "
                    "'{}' used, '{}' expected.".format(
                        prefix_name, prefix, str(net)
                    )
                )

        cls.info("{}: setting instances up...".format(cls.SHORT_DESCR))

        MockedEnv(cls, base_dir=cls._get_module_dir(),
                  peering_db=cls.MOCK_PEERING_DB,
                  ripe_rpki_cache=cls.MOCK_RIPE_RPKI_CACHE,
                  irrdb=cls.MOCK_IRRDB,
                  rttgetter=cls.MOCK_RTTGETTER,
                  arin_db_dump=cls.MOCK_ARIN_DB_DUMP)

        cls.rs_cfg_file_path = None

        try:
            cls._setup_instances()
        except:
            cls.tearDownClass()
            raise

        if "BUILD_ONLY" in os.environ or \
            (cls.SKIP_ON_TRAVIS and "TRAVIS" in os.environ):
            cls.debug("Skipping starting instances")
            return

        try:
            for instance in cls.INSTANCES:
                instance.set_var_dir("{}/var".format(cls._get_module_dir()))

                if cls._do_not_stop_instances() and instance.is_running():
                    cls.debug("Instance '{}' already running, reloading config".format(instance.name))
                    if not instance.reload_config():
                        raise InstanceError("An error occurred while reloading '{}' configuration.".format(instance.name))
                    continue

                cls.debug("Starting instance '{}'...".format(instance.name))
                instance.start()
        except Exception as e:
            try:
                cls.tearDownClass()
            except:
                pass
            raise e

    @staticmethod
    def get_instance_tag(instance_or_class):
        if isinstance(instance_or_class, BGPSpeakerInstance):
            _class = instance_or_class.__class__
        else:
            _class = instance_or_class

        try:
            tag = _class.TAG
        except:
            msg = "'TAG' missing for instance of type "
            if isinstance(instance_or_class, BGPSpeakerInstance):
                msg += instance_or_class.name
            else:
                msg += instance_or_class.__name__
            raise ValueError(msg)

        return tag

    @classmethod
    def dump_routes(cls):
        rs_tag = cls.get_instance_tag(cls.RS_INSTANCE_CLASS)

        dest_dir = os.path.join(cls._get_module_dir(), "routes", cls.__name__, rs_tag)
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)

        for instance in cls.INSTANCES:
            if not isinstance(instance, BGPSpeakerInstance):
                continue

            path = os.path.join(dest_dir, "{}.txt".format(instance.name))
            routes = instance.get_routes(None, include_filtered=True)
            sorted_routes = sorted(routes,
                                   key = lambda route: (route.prefix,
                                                        route.as_path,
                                                        route.next_hop,
                                                        route.via))
            with open(path, "w") as f:
                for route in sorted_routes:
                    route.dump(f)

    @classmethod
    def dump_rs_config(cls):
        if not cls.rs_cfg_file_path:
            return

        dest_dir = os.path.join(cls._get_module_dir(), "configs", cls.__name__)
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)

        tag = cls.get_instance_tag(cls.RS_INSTANCE_CLASS)

        path = os.path.join(dest_dir, "{}.conf".format(tag))
        with open(path, "w") as dst:
            with open(cls.rs_cfg_file_path, "r") as src:
                dst.write(src.read())

    @classmethod
    def _tearDownClass(cls):
        MockedEnv.stopall()

        if "TRAVIS" not in os.environ:
            cls.info("{}: dumping rs config...".format(cls.SHORT_DESCR))
            cls.dump_rs_config()

        if "BUILD_ONLY" in os.environ or \
            (cls.SKIP_ON_TRAVIS and "TRAVIS" in os.environ):
            cls.debug("Skipping stopping instances")
            return

        if "TRAVIS" not in os.environ:
            cls.info("{}: dumping routes...".format(cls.SHORT_DESCR))
            cls.dump_routes()

        if cls._do_not_stop_instances():
            cls.debug("Skipping stopping instances")
            return

        cls.info("{}: stopping instances...".format(cls.SHORT_DESCR))

        for instance in cls.INSTANCES:
            cls.debug("Stopping instance '{}'...".format(instance.name))
            instance.stop()

    def set_instance_variables(self):
        raise NotImplementedError()

    def _setUp(self):
        if "BUILD_ONLY" in os.environ:
            self.skipTest("Build only")

        if self.SKIP_ON_TRAVIS and "TRAVIS" in os.environ:
            self.skipTest("not supported on Travis CI")

        self.set_instance_variables()

    def process_reject_cause_routes(self, routes):
        if self.REJECT_CAUSE_COMMUNITY is not None:
            re_pattern = re.compile(self.REJECT_CAUSE_COMMUNITY)
            rejected_route_announced_by_pattern = None
            if self.REJECTED_ROUTE_ANNOUNCED_BY_COMMUNITY:
                rejected_route_announced_by_pattern = \
                    re.compile(self.REJECTED_ROUTE_ANNOUNCED_BY_COMMUNITY)

            for route in routes:
                route.process_reject_cause(re_pattern,
                                           rejected_route_announced_by_pattern)

    def _instance_log_contains_errors_warning(self, inst):
        if inst.log_contains_errors(self.ALLOWED_LOG_ERRORS):
            return "\nWARNING: {} log contains errors".format(inst.name)
        return ""

    def receive_route(self, inst, prefix, other_inst=None, as_path=None,
                      next_hop=None, std_comms=None, lrg_comms=None,
                      ext_comms=None, local_pref=None, as_set=None, otc=None,
                      filtered=None, only_best=None, reject_reason=None):
        """Test if the BGP speaker receives the expected route(s).

        If no routes matching the given criteria are found, the
        ``TestCase.fail()`` method is called and the test fails.

        Args:
            inst: the :class:`BGPSpeakerInstance` instance where the routes
                are searched on.

            prefix (str): the IPv4/IPv6 prefix of the routes to search for.

            other_inst: if given, only routes received from this
                :class:`BGPSpeakerInstance` instance are considered.

            as_path (str): if given, only routes with this AS_PATH are
                considered.

            next_hop: can be a string or a :class:`BGPSpeakerInstance`
                instance; if given, only routes that have a NEXT_HOP
                address matching this one are considered.

            std_comms, lrg_comms, ext_comms (list): if given, only routes
                that carry these BGP communities are considered. Use an
                empty list ([]) to consider only routes with no BGP comms.

            local_pref (int): if given, only routes with local-pref equal
                to this value are considered.

            as_set (str): if given, only routes with this AS_SET are
                considered.

            otc (int): if provided, only routes with the OTC attribute set
                to this value are considered. Use '0' to match only routes
                NOT having the OTC value set.

            filtered (bool): if given, only routes that have been (not)
                filtered are considered.

            only_best (bool): if given, only best routes are considered.

            reject_reason (int): valid only if `filtered` is True: if given
                the route must be reject with this reason code.
                It can be also a set of codes: in this case, the route must
                be rejected with one of those codes.

                The list of valid codes is reported in docs/CONFIG.rst or at
                https://arouteserver.readthedocs.io/en/latest/CONFIG.html#reject-policy
        """
        assert isinstance(inst, BGPSpeakerInstance), \
            "inst must be of class BGPSpeakerInstance"

        try:
            IPNetwork(prefix)
        except:
            raise AssertionError("prefix must be a valid IPv4/IPv6 prefix")

        if other_inst:
            assert isinstance(other_inst, BGPSpeakerInstance), \
                "other_inst must be of class BGPSpeakerInstance"

        if as_path:
            try:
                if not isinstance(as_path, str):
                    raise AssertionError()
                for asn in as_path.split(" "):
                    assert asn.strip().isdigit()
            except:
                raise AssertionError("as_path must be a string in the format "
                                     "'<asn1> <asn2>' with <asnX> positive "
                                     "integers")

        next_hop_ip = None
        if next_hop:
            assert isinstance(next_hop, (str, BGPSpeakerInstance)), \
                ("next_hop must be a string representing one IP address or "
                 "a BGPSpeakerInstance object")
            next_hop_ip = next_hop if isinstance(next_hop, str) else next_hop.ip
            try:
                IPAddress(next_hop_ip)
            except:
                raise AssertionError("Invalid next_hop IP address: {}".format(
                    next_hop_ip
                ))

        if std_comms:
            assert isinstance(std_comms, list), \
                ("std_comms must be a list of strings representing "
                 "BGP standard communities")

        if ext_comms:
            assert isinstance(ext_comms, list), \
                ("ext_comms must be a list of strings representing "
                 "BGP extended communities")

        if lrg_comms:
            assert isinstance(lrg_comms, list), \
                ("lrg_comms must be a list of strings representing "
                 "BGP large communities")

        if local_pref:
            assert isinstance(local_pref, int), \
                "local_pref must be an integer >= 0"
            assert local_pref >= 0, \
                "local_pref must be an integer >= 0"

        if otc is not None:
            assert isinstance(otc, int), \
                "otc must be an integer"
            assert otc >= 0, \
                "otc must be an integer >= 0"

        if reject_reason is not None and not filtered:
            raise AssertionError(
                "reject_reason can be set only if filtered is True"
            )

        reject_reasons = None
        if reject_reason is not None:
            if isinstance(reject_reason, int):
                reject_reasons = [reject_reason]
            else:
                reject_reasons = list(reject_reason)
            for code in reject_reasons:
                assert code in range(1,16), "invalid reject_reason"

        include_filtered = filtered if filtered is not None else False
        best_only = only_best if only_best is not None else False

        routes = inst.get_routes(prefix,
                                 include_filtered=include_filtered,
                                 only_best=best_only)
        self.process_reject_cause_routes(routes)

        errors = []
        if not routes:
            errors.append("{inst} does not receive {prefix} at all.")
        else:
            for route in routes:
                err = False
                if other_inst and route.via != other_inst.ip:
                    errors.append("{{inst}} receives {{prefix}} from {via} and not from {{other_inst}}.".format(via=route.via))
                    err = True
                if as_path and route.as_path != as_path:
                    errors.append("{{inst}} receives {{prefix}} with AS_PATH {as_path} and not with {{as_path}}.".format(as_path=route.as_path))
                    err = True
                if next_hop_ip and route.next_hop != next_hop_ip:
                    errors.append("{{inst}} receives {{prefix}} with NEXT_HOP {next_hop} and not with {{next_hop_ip}}.".format(next_hop=route.next_hop))
                    err = True
                if std_comms is not None and sorted(route.std_comms) != sorted(std_comms):
                    errors.append("{{inst}} receives {{prefix}} with std comms {comms} and not with {{std_comms}}.".format(comms=route.std_comms))
                    err = True
                if lrg_comms is not None and sorted(route.lrg_comms) != sorted(lrg_comms):
                    errors.append("{{inst}} receives {{prefix}} with lrg comms {comms} and not with {{lrg_comms}}.".format(comms=route.lrg_comms))
                    err = True
                if ext_comms is not None and sorted(route.ext_comms) != sorted(ext_comms):
                    errors.append("{{inst}} receives {{prefix}} with ext comms {comms} and not with {{ext_comms}}.".format(comms=route.ext_comms))
                    err = True
                if local_pref is not None and route.localpref != local_pref:
                    errors.append("{{inst}} receives {{prefix}} with local-pref {local_pref} and not with {{local_pref}}.".format(local_pref=route.localpref))
                    err = True
                if as_set is not None and route.as_set != as_set:
                    errors.append("{{inst}} receives {{prefix}} with AS_SET {as_set} and not with {{as_set}}.".format(as_set=route.as_set))
                    err = True
                if otc is not None and (
                    (
                        otc > 0 and route.otc != otc
                    ) or (
                        otc == 0 and route.otc is not None
                    )
                ):
                    errors.append("{{inst}} receives {{prefix}} with otc {otc} and not with {{otc}}.".format(otc=route.otc))
                    err = True
                if filtered is not None and route.filtered != filtered:
                    errors.append(
                        "{{inst}} receives {{prefix}} from {via}, AS_PATH {as_path}, NEXT_HOP {next_hop} "
                        "but it is {filtered_status} while it is expected to be {filtered_exp}.".format(
                            via=route.via,
                            as_path=route.as_path if not route.as_set else route.as_path + " (AS_SET {})".format(route.as_set),
                            next_hop=route.next_hop,
                            filtered_status="filtered" if route.filtered else "not filtered",
                            filtered_exp="filtered" if filtered else "not filtered"
                        )
                    )
                    err = True
                if filtered is True and route.filtered and \
                    reject_reasons is not None and len(route.reject_reasons) > 0:

                    reject_reason_found = False
                    for real_reason in route.reject_reasons:
                        if real_reason in reject_reasons:
                            reject_reason_found = True

                    if not reject_reason_found:
                        if len(reject_reasons) == 1:
                            exp_reason = reject_reasons[0]
                        else:
                            exp_reason = "one of {}".format(", ".join(map(str, reject_reasons)))

                        errors.append(
                            "{{inst}} receives {{prefix}} from {via}, AS_PATH {as_path}, NEXT_HOP {next_hop}, "
                            "it is filtered but reject reasons don't match: real reasons {reason}, "
                            "expected reason {exp_reason}.".format(
                                via=route.via,
                                as_path=route.as_path,
                                next_hop=route.next_hop,
                                reason=", ".join(map(str, route.reject_reasons)),
                                exp_reason=exp_reason
                            )
                        )
                        err = True
                if not err:
                    return route

        if errors:
            criteria = []
            if other_inst:
                criteria.append("from {} ({})".format(other_inst.ip, other_inst.name))
            if as_path:
                criteria.append("with AS_PATH {}".format(as_path))
            if next_hop_ip:
                criteria.append("with next-hop {}".format(next_hop_ip))
            if std_comms:
                criteria.append("with std comms {}".format(std_comms))
            if lrg_comms:
                criteria.append("with lrg comms {}".format(lrg_comms))
            if ext_comms:
                criteria.append("with ext comms {}".format(ext_comms))
            if local_pref:
                criteria.append("with local-pref {}".format(local_pref))
            if as_set:
                criteria.append("with as_set {}".format(as_set))
            if otc:
                criteria.append("with OTC {}".format(otc))
            if filtered is True:
                criteria.append("filtered")
            if reject_reasons:
                if len(reject_reasons) == 1:
                    criteria.append(
                        "with reject reason {}".format(reject_reasons[0])
                    )
                else:
                    criteria.append(
                        "with reject reason in {}".format(
                            ", ".join(map(str, reject_reasons))
                        )
                    )

            failure = "Routes not found.\n"
            failure += "Looking for {prefix} on {inst} {criteria}:\n\t".format(
                prefix=prefix,
                inst=inst.name,
                criteria=", ".join(criteria)
            )
            failure += "\n\t".join([
                err_msg.format(
                    inst=inst.name,
                    prefix=prefix,
                    other_inst="{} ({})".format(other_inst.ip, other_inst.name) if other_inst else "",
                    as_path=as_path,
                    next_hop_ip=next_hop_ip,
                    std_comms=std_comms,
                    lrg_comms=lrg_comms,
                    ext_comms=ext_comms,
                    local_pref=local_pref,
                    otc=otc,
                    as_set=as_set,
                ) for err_msg in errors
            ])
            failure += self._instance_log_contains_errors_warning(inst)
            self.fail(failure)

    def log_contains(self, inst, msg, instances={}, opposite=False):
        """Test if the BGP speaker's log contains the expected message.

        This only works for BGP speaker instances that support message
        logging: currently only BIRD.

        If no log entries are found, the ``TestCase.fail()`` method is
        called and the test fails. If ``opposite`` is ``True``, the
        failure is reported if a log entry is found.

        Args:

            inst: the :class:`BGPSpeakerInstance` instance where the
                expected message is searched on.

            msg (str): the text that is expected to be found within
                BGP speaker's log.

            instances (dict): a dictionary of pairs
                "<macro>: <BGPSpeakerInstance>" used to expand macros on
                the *msg* argument. Macros are expanded using the BGP
                speaker's specific client ID or protocol name.

            opposite (bool): when set to True, the call fails if a match
                is found.

        Example
        ---------

        Given *self.rs* the instance of the route server, and *self.AS1* the
        instance of one of its clients, the following code expands the "{AS1}"
        macro using the BGP speaker specific name for the instance *self.AS1*
        and then looks for it within the route server's log:

            ``self.log_contains(self.rs, "{AS1} bad ASN", {"AS1": self.AS1})``

        On BIRD, "{AS1}" will be expanded using the "protocol name" that BIRD
        uses to identify the BGP session with AS1.
        """
        if not inst.MESSAGE_LOGGING_SUPPORT:
            return

        expanded_msg = msg

        macros_dict = {}
        for inst_name in instances:
            other_inst = instances[inst_name]
            # BAD THING: using a BIRD specific method to understand
            # what's the other instance's name on 'inst'.
            proto_name = inst.get_protocol_name_by_ip(other_inst.ip)
            macros_dict[inst_name] = proto_name

        if macros_dict:
            expanded_msg = expanded_msg.format(**macros_dict)

        if not opposite and not inst.log_contains(expanded_msg):
            self.fail(
                "Expected message not found on {} logs:\n\t{}".format(
                    inst.name, expanded_msg
                ) + self._instance_log_contains_errors_warning(inst)
            )

        if opposite and inst.log_contains(expanded_msg):
            self.fail(
                "Unexpected message found on {} logs:\n\t{}".format(
                    inst.name, expanded_msg
                ) + self._instance_log_contains_errors_warning(inst)
            )

    def session_exists(self, inst_a, inst_b_or_ip):
        """Test if a BGP session between the two instances exists.

        Args:
            inst_a: the :class:`BGPSpeakerInstance` instance where the
                BGP session is looked for.

            inst_b_or_ip: the :class:`BGPSpeakerInstance` instance or an
                IP address that *inst_a* is expected to peer with.
        """

        if inst_a.get_bgp_session(inst_b_or_ip) is None:
            self.fail(
                "A BGP session between '{}' ({}) and '{}' "
                "does not exist.".format(
                    inst_a.name, inst_a.ip,
                    "{} ({})".format(
                        inst_b_or_ip.name, inst_b_or_ip.ip
                    ) if isinstance(inst_b_or_ip, BGPSpeakerInstance)
                    else inst_b_or_ip
                ) + self._instance_log_contains_errors_warning(inst_a)
            )

    def session_is_up(self, inst_a, inst_b):
        """Test if a BGP session between the two instances is up.

        If a BGP session between the two instances is not up, the
        ``TestCase.fail()`` method is called and the test fails.

        Args:
            inst_a: the :class:`BGPSpeakerInstance` instance where the
                BGP session is looked for.

            inst_b: the :class:`BGPSpeakerInstance` instance that *inst_a* is
                expected to peer with.
        """

        self.session_exists(inst_a, inst_b)

        for _ in range(5):
            if inst_a.bgp_session_is_up(inst_b):
                return
            time.sleep(1)
        for _ in range(20):
            if inst_a.bgp_session_is_up(inst_b, force_update=True):
                return
            time.sleep(1)
        self.fail(
            "BGP session between '{}' ({}) and '{}' ({}) is not up.".format(
                inst_a.name, inst_a.ip, inst_b.name, inst_b.ip
            ) + self._instance_log_contains_errors_warning(inst_a)
        )

    def test_010_setup(self):
        # Referenced by utils/update_tests, used when BUILD_ONLY=1
        raise NotImplementedError()

    def test_999_log_contains_errors(self):
        """{}: log contains errors"""
        try:
            errors_found, errors = self.rs.log_contains_errors(
                self.ALLOWED_LOG_ERRORS, True
            )
            if errors_found:
                self.fail("rs log contains errors:\n{}".format(errors))
        except AttributeError:
            raise NotImplementedError("self.rs does not exist here")
        except NotImplementedError:
            raise

class LiveScenario_TagRejectPolicy(object):
    """Helper class to run a scenario as if reject_policy is set to 'tag'.

    When a scenario inherits this class, its route server is configured as
    if the ``reject_policy.policy`` is ``tag``, the ``65520:dyn_val``
    value is used for the ``reject_cause`` BGP community and the
    ``65524:dyn_val`` and ``rt:65524:dyn_val`` values for the
    ``rejected_route_announced_by`` one.
    Additionally, using the ``reject_cause_map``, some reject codes are
    mapped to specific BGP communities in the ``rs_as:1101:*`` range.

    The ``general.yml`` file, or the file given in the ``orig_file`` argument
    of ``_get_cfg_general`` method, is cloned and reconfigured with the
    aforementioned settings.

    This class is mostly used for OpenBGPD tests since the underlaying
    mechanism allows to track the reason that brought to consider the route as
    rejected and to test filters out during test cases execution.

    This class should be used in multiple inheritance:

    Example
    ---------

        ``class SkeletonScenario_OpenBGPDIPv4(LiveScenario_TagRejectPolicy, SkeletonScenario):``

    """

    REJECT_CAUSE_COMMUNITY = r"^65520:(\d+)$"
    REJECTED_ROUTE_ANNOUNCED_BY_COMMUNITY = r"^rt:65524:(\d+)$"
    REJECT_POLICY = "tag"

    @classmethod
    def _get_cfg_general(cls, orig_file="general.yml"):
        orig_path = "{}/{}".format(cls._get_module_dir(), orig_file)
        dest_rel_path = "var/general.yml"
        dest_path = "{}/{}".format(cls._get_module_dir(), dest_rel_path)

        with open(orig_path, "r") as f:
            cfg = yaml.safe_load(f.read())

        cfg["cfg"]["filtering"]["reject_policy"] = {"policy": cls.REJECT_POLICY}
        if "communities" not in cfg["cfg"]:
            cfg["cfg"]["communities"] = {}
        cfg["cfg"]["communities"]["reject_cause"] = {"std": "65520:dyn_val"}
        cfg["cfg"]["communities"]["rejected_route_announced_by"] = {
            "std": "65524:dyn_val",
            "ext": "rt:65524:dyn_val"
        }
        cfg["cfg"]["communities"]["reject_cause_map"] = {
            6: {
                "lrg": "rs_as:1101:7"
            }
        }

        with open(dest_path, "w") as f:
            yaml.safe_dump(cfg, f, default_flow_style=False)

        return dest_rel_path

class LiveScenario_TagAndRejectRejectPolicy(LiveScenario_TagRejectPolicy):
    """Same as LiveScenario_TagRejectPolicy, but with 'tag_and_reject' reject policy."""

    REJECT_POLICY = "tag_and_reject"



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import BaseConfigEnricher
from ..errors import BuilderError, PeeringDBError, PeeringDBNoInfoError
from ..peering_db import PeeringDBNetNeverViaRouteServers

class NeverViaRouteServersEnricher(BaseConfigEnricher):

    def enrich(self):

        logging.info("Retrieving 'never via route-servers' networks from PeeringDB...")

        try:
            peeringdb_data = PeeringDBNetNeverViaRouteServers(
               cache_dir=self.builder.cache_dir,
               cache_expiry=self.builder.cache_expiry,
            )
            peeringdb_data.load_data()
        except PeeringDBNoInfoError:
            # No data found on PeeringDB.
            logging.warning("No networks found on PeeringDB "
                            "with 'info_never_via_route_servers' "
                            "attribute set.")
            return
        except PeeringDBError as e:
            logging.error(
                "An error occurred while retrieving 'never via route-servers' "
                "networks from PeeringDB: {}".format(
                    str(e) or "error unknown"
                )
            )
            raise BuilderError()

        for network in peeringdb_data.networks:
            asn = network["asn"]
            if asn not in self.builder.never_via_route_servers_asns:
                self.builder.never_via_route_servers_asns.append(asn)

        logging.info("{} 'never via route-servers' networks "
                     "found on PeeringDB".format(len(peeringdb_data.networks)))



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
REJECT_REASONS = {
"1": "Invalid AS_PATH length",
"2": "Prefix is bogon",
"3": "Prefix is in global blacklist",
"4": "Invalid AFI",
"5": "Invalid NEXT_HOP",
"6": "Invalid left-most ASN",
"7": "Invalid ASN in AS_PATH",
"8": "Transit-free ASN in AS_PATH",
"9": "Origin ASN not in IRRDB AS-SETs",
"10": "IPv6 prefix not in global unicast space",
"11": "Prefix is in client blacklist",
"12": "Prefix not in IRRDB AS-SETs",
"13": "Invalid prefix length",
"14": "RPKI INVALID route",
"15": "Never via route-servers ASN in AS_PATH",
}



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import atexit
import logging
import shutil
import pickle
import tempfile

from .base import BaseConfigEnricher, BaseConfigEnricherThread
from ..errors import BuilderError, ARouteServerError
from ..ipaddresses import IPAddress, IPNetwork
from ..irrdb import ASSet, RSet, AS_SET_Bundle


def clear_irrdb_pickle_dir(target_dir):
    shutil.rmtree(target_dir, ignore_errors=True)

class IRRDB(object):
    """
    Container for IRRDB "records".
    Each record represents a "bundle" of one or more objects and the
    data resulting from their expansion (prefixes and/or ASNs).
    A bundle is made up by one or more aut-num/as-set objects that are
    passed to bgpq3 to retrieve prefixes and/or ASNs; bgpq3's ability
    to aggregate prefixes is used to merge the content of each object
    into a single dataset.
    """

    def __init__(self):
        self.irrdb_pickle_dir = tempfile.mkdtemp(suffix="_arouteserver")
        atexit.register(clear_irrdb_pickle_dir,
                        target_dir=self.irrdb_pickle_dir)

        self.records = {}

    def request(self, names, used_by, object_types=set(["prefixes", "asns"])):
        assert object_types.issubset(set(["prefixes", "asns"]))

        names_list = names if isinstance(names, list) else [names]

        used_source, same_for_all = AS_SET_Bundle.get_source(names_list)
        if not same_for_all:
            logging.info(
                "IRRDB: the source of the bundle '{}' "
                "used by {} is not the same for all the items: "
                "{} will be used for all".format(
                    ", ".join(names_list), used_by,
                    used_source if used_source else "the default sources"
                )
            )

        new_record = IRRDBRecord(names_list, self.irrdb_pickle_dir)

        if new_record.id not in self.records:
            self.records[new_record.id] = new_record
        record = self.records[new_record.id]

        record.used_by.add(used_by)
        record.requested_objects.update(object_types)

        return record.id

    def __getitem__(self, name):
        return self.records[name]

    def __iter__(self):
        return iter(self.records)

    def keys(self):
        return self.records.keys()

    def items(self):
        return self.records.items()

    def iteritems(self):
        return self.records.items()

    def values(self):
        return self.records.values()

class IRRDBRecord(AS_SET_Bundle):

    def __init__(self, as_set_names, irrdb_pickle_dir):
        AS_SET_Bundle.__init__(self, as_set_names)

        self.used_by = set()

        self.requested_objects = set()

        self.irrdb_pickle_dir = irrdb_pickle_dir

        self.saved_objects = []

    def get_path(self, objects):
        return "{}/pickle_{}.{}".format(self.irrdb_pickle_dir, self.id, objects)

    def save(self, objects, data):
        if objects in ("asns", "prefixes"):
            with open(self.get_path(objects), "wb") as f:
                pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)
            self.saved_objects += [objects]
        else:
            raise ValueError("Unknown objects: {}".format(objects))

    def load(self, objects):
        if objects in self.saved_objects:
            with open(self.get_path(objects), "rb") as f:
                return pickle.load(f)
        else:
            return []

    @property
    def asns(self):
        return self.load("asns")

    @property
    def prefixes(self):
        return self.load("prefixes")

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "descr": self.descr,
            "used_by": ", ".join(sorted(self.used_by)),
            "asns": self.asns,
            "prefixes": self.prefixes
        }

class IRRDBConfigEnricher_WorkerThread(BaseConfigEnricherThread):

    TARGET_FIELD = None

    def __init__(self, *args, **kwargs):
        BaseConfigEnricherThread.__init__(self, *args, **kwargs)

        self.ip_ver = None
        self.irrdbtools_cfg = None

    def do_task(self, task):
        irrdb_record = task
        used_by_descr = ", ".join(irrdb_record.used_by)
        data = self._get_external_data(used_by_descr, irrdb_record.object_names)
        return data

    def save_data(self, task, data):
        irrdb_record = task
        irrdb_record.save(self.TARGET_FIELD, data)

class IRRDBConfigEnricher_WorkerThread_Prefixes(IRRDBConfigEnricher_WorkerThread):

    DESCR = "IRRdb prefixes"
    TARGET_FIELD = "prefixes"

    def _get_external_data(self, used_by_descr, as_set_names):
        errors = False
        ip_versions = [self.ip_ver] if self.ip_ver else [4, 6]
        res = []
        for ip_ver in ip_versions:
            obj = RSet(as_set_names, ip_ver, **self.irrdbtools_cfg)
            try:
                obj.load_data()
                prefixes = obj.prefixes
                if not prefixes:
                    logging.warning("No IPv{} prefixes found in "
                                    "{} for {}".format(
                                        ip_ver, obj.descr, used_by_descr))
                res.extend(prefixes)
            except ARouteServerError as e:
                errors = True
                logging.error(
                    "Error while retrieving IPv{} prefixes "
                    "from {} for {}: {}".format(
                        ip_ver, obj.descr, used_by_descr, str(e)
                    )
                )

        if errors:
            raise BuilderError()

        return res

class IRRDBConfigEnricher_WorkerThread_ASNs(IRRDBConfigEnricher_WorkerThread):

    DESCR = "IRRdb origin ASNs"
    TARGET_FIELD = "asns"

    def _get_external_data(self, used_by_descr, as_set_name):
        errors = False
        obj = ASSet(as_set_name, **self.irrdbtools_cfg)
        try:
            obj.load_data()
            asns = obj.asns
            if not asns:
                logging.warning("No origin ASNs found in "
                                "{} for {}".format(
                                    obj.descr, used_by_descr))
            return asns
        except ARouteServerError as e:
            errors = True
            logging.error(
                "Error while retrieving origin ASNs from {} for {}: {}".format(
                    obj.descr, used_by_descr, str(e)
                )
            )

        if errors:
            raise BuilderError()

class IRRDBConfigEnricher(BaseConfigEnricher):

    WORKER_THREAD_CLASS = None

    WHITE_LIST_OBJECT_NAME_PREFIX = "WHITE_LIST_"

    def prepare(self):
        # Create and populate the IRRDB() instances only once
        # (two IRR enrichers are used, but only the first one builds it).
        if self.builder.irrdb_info is not None:
            return

        # In the end, self.builder.irrdb_info will be an IRRDB()
        # object. Its items will be:
        # {
        #   "<as_set_bundle_id>": <IRRDBRecord>
        # }

        self.builder.irrdb_info = IRRDB()

        # Add to irrdb_info all the AS-SET bundles reported in the 'clients' section.
        for client in self.builder.cfg_clients.cfg["clients"]:
            client_irrdb = client["cfg"]["filtering"]["irrdb"]
            client_irrdb["as_set_bundle_ids"] = set()

            if not client_irrdb["enforce_origin_in_as_set"] and \
                not client_irrdb["enforce_prefix_in_as_set"] and \
                not self.builder.cfg_general["filtering"]["irrdb"]["tag_as_set"]:

                # Client does not require AS-SETs info to be gathered.
                continue

            if self.builder.ip_ver is not None:
                ip = client["ip"]
                if IPAddress(ip).version != self.builder.ip_ver:
                    # The address family of this client is not the
                    # current one used to build the configuration.
                    continue

            # Client needs AS-SETs info because origin ASN or prefix filters
            # are required.

            # In the worst case, use AS<asn>.
            client_irrdb["as_set_bundle_ids"].add(
                self.builder.irrdb_info.request(
                    "AS{}".format(client["asn"]),
                    "client {}".format(client["id"])
                )
            )

            # IRR white lists
            for cfg_attr, obj_type in (("white_list_pref", "prefixes"),
                                       ("white_list_asn", "asns")):

                white_list_objects = client_irrdb[cfg_attr]

                if not white_list_objects:
                    continue

                if obj_type == "prefixes" and self.builder.ip_ver:
                    # Only consider prefixes for the current IP version.
                    ip_ver = self.builder.ip_ver
                    white_list_objects = [
                        p for p in white_list_objects
                        if IPNetwork(p["prefix"]).version == ip_ver
                    ]

                if white_list_objects:
                    # If a white list of prefixes/ASNs has been set for the
                    # client, add a fake 'white_list' AS-SET with those
                    # prefixes/ASNs.
                    white_list_name = "{prefix}{client_id}".format(
                        prefix=self.WHITE_LIST_OBJECT_NAME_PREFIX,
                        client_id=client["id"]
                    )
                    white_list_record_id = self.builder.irrdb_info.request(
                        white_list_name,
                        "client {} white list".format(client["id"]),
                        set()
                    )
                    self.builder.irrdb_info[white_list_record_id].save(
                        obj_type, client_irrdb[cfg_attr]
                    )
                    if white_list_record_id not in client_irrdb["as_set_bundle_ids"]:
                        client_irrdb["as_set_bundle_ids"].add(white_list_record_id)

            # Client has its own specific set of AS-SETs.
            if client_irrdb["as_sets"]:
                client_irrdb["as_set_bundle_ids"].add(
                    self.builder.irrdb_info.request(
                        client_irrdb["as_sets"],
                        "client {}".format(client["id"])
                    )
                )
                continue

            # Client needs AS-SETs info but has not its own set of AS-SETs.

            # If client's ASN is configured in the 'asns' section and has
            # one or more AS-SETs, the client configuration will be based
            # on those.
            asn = "AS{}".format(client["asn"])
            if asn in self.builder.cfg_asns.cfg["asns"] and \
                self.builder.cfg_asns.cfg["asns"][asn]["as_sets"]:
                client_irrdb["as_set_bundle_ids"].add(
                    self.builder.irrdb_info.request(
                        self.builder.cfg_asns.cfg["asns"][asn]["as_sets"],
                        "client {}".format(client["id"])
                    )
                )
                continue

            # If one or more AS-SETs have been found on PeeringDB,
            # use them.
            as_sets_from_pdb = client.get("as_sets_from_pdb", None)
            if as_sets_from_pdb:
                logging.info("No AS-SETs provided for the '{}' client. "
                             "Using AS{} + those obtained from PeeringDB: "
                             "{}.".format(
                                    client["id"], client["asn"],
                                    ", ".join(as_sets_from_pdb)
                                ))
                client_irrdb["as_set_bundle_ids"].add(
                    self.builder.irrdb_info.request(
                        as_sets_from_pdb,
                        "client {}".format(client["id"])
                    )
                )
                continue

            # No other AS-SETs found for the client.
            logging.warning("No AS-SETs provided for the '{}' client. "
                            "Only AS{} will be expanded.".format(
                                client["id"], client["asn"]
                            ))

    def _config_thread(self, thread):
        thread.ip_ver = self.builder.ip_ver
        thread.irrdbtools_cfg = {
            "bgpq3_path": self.builder.bgpq3_path,
            "bgpq3_host": self.builder.bgpq3_host,
            "bgpq3_sources": self.builder.bgpq3_sources,
            "bgpq3_timeout": self.builder.bgpq3_timeout,
            "cache_dir": self.builder.cache_dir,
            "cache_expiry": self.builder.cache_expiry,
        }

    def add_tasks(self):
        target_objects = self.WORKER_THREAD_CLASS.TARGET_FIELD

        # Enqueuing tasks.
        for _, as_set_record in self.builder.irrdb_info.items():
            if target_objects in as_set_record.requested_objects:
                self.tasks_q.put(as_set_record)

class IRRDBConfigEnricher_ASNs(IRRDBConfigEnricher):

    WORKER_THREAD_CLASS = IRRDBConfigEnricher_WorkerThread_ASNs

class IRRDBConfigEnricher_Prefixes(IRRDBConfigEnricher):

    WORKER_THREAD_CLASS = IRRDBConfigEnricher_WorkerThread_Prefixes

    def _config_thread(self, thread):
        IRRDBConfigEnricher._config_thread(self, thread)
        thread.irrdbtools_cfg["allow_longer_prefixes"] = \
            self.builder.cfg_general["filtering"]["irrdb"]["allow_longer_prefixes"]



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import SkeletonScenario
from pierky.arouteserver.builder import BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.bird import BIRDInstanceIPv4

class SkeletonScenario_BIRDIPv4(SkeletonScenario):
    """BGP speaker specific and IP version specific derived class.

    This class inherits all the test functions from the base class.
    Here, only IP version specific attributes are set, such as the
    prefix IDs / real IP prefixes mapping schema.

    The prefix IDs reported within the ``DATA`` dictionary must be
    used in the parent class' test functions to reference the real
    IP addresses/prefixes used in the scenario. Also the other
    BGP speakers' configuration templates must use these IDs.
    For an example plase see the "AS2.j2" file.

    The ``SHORT_DESCR`` attribute can be set with a brief description
    of this scenario.
    """

    # Leave this to True in order to allow pytest to use this class
    # to run tests.
    __test__ = True

    SHORT_DESCR = "Live test, BIRD, skeleton, IPv4"
    CONFIG_BUILDER_CLASS = BIRDConfigBuilder
    RS_INSTANCE_CLASS = BIRDInstanceIPv4
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv4
    IP_VER = 4

    DATA = {
        "rs_IPAddress":             "192.0.2.2",
        "AS1_IPAddress":            "192.0.2.11",
        "AS2_IPAddress":            "192.0.2.22",

        "AS2_prefix1":              "2.0.1.0/24",
        "AS2_bogon1":               "192.168.2.0/24"
    }



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import datetime
import sys
import yaml

from .base import ARouteServerCommand
from ..config.program import program_config
from ..config.clients import merge_clients
from ..euro_ix import EuroIXMemberList
from ..errors import ARouteServerError

class ClientsFromEuroIXCommand(ARouteServerCommand):

    COMMAND_NAME = "clients-from-euroix"
    COMMAND_HELP = ("Build a list of clients on the basis "
                    "of EURO-IX JSON file.")
    NEEDS_CONFIG = True

    @classmethod
    def add_arguments(cls, parser):
        super(ClientsFromEuroIXCommand, cls).add_arguments(parser)

        parser.add_argument(
            "--url",
            type=str,
            help="URL of the file that contains the Euro-IX JSON "
                 "member list. If this is not given, data is read "
                 "from the input file (arguments -i, --input).",
            dest="url")

        parser.add_argument(
            "-i", "--input",
            type=argparse.FileType('r'),
            help="Input file. Default: stdin.",
            default=sys.stdin,
            dest="input_file")

        parser.add_argument(
            "ixp_id",
            type=int,
            nargs="?",
            help="The numeric identifier used by the IX to identify the "
                 "infrastructure for which the list of clients is requested. "
                 "If not given, a list of IX's infrastructures will be "
                 "printed.")

        parser.add_argument(
            "--vlan-id",
            type=int,
            help="Only consider members that have a connection to the given "
                 "VLAN id. Please note: this is not necessarily the 802.1q "
                 "tag but, generally speaking, it is the numeric identifier "
                 "used by the IX to identify the VLAN in its Euro-IX JSON "
                 "file.",
            dest="vlan_id")

        parser.add_argument(
            "--routeserver-only",
            action="store_true",
            help="Exclude members whose "
                 "'ixp_list.[].vlan.[].ipv[4|6].routeserver' attribute "
                 "is false.",
            dest="routeserver_only")

        parser.add_argument(
            "--guess-custom-bgp-communities",
            nargs="+",
            choices=EuroIXMemberList.CUSTOM_COMMUNITIES,
            help="If set, clients will be configured to attach custom "
                 "informational BGP communities to the routes they "
                 "announce to the route server. These communities will "
                 "be guessed on the basis of the available attributes "
                 "from the Euro-IX JSON file (member type, switch ID, "
                 "city, ...).",
            dest="guess_custom_bgp_communities")


        parser.add_argument(
            "--merge-from-peeringdb",
            nargs="+",
            choices=EuroIXMemberList.INFO_FROM_PEERINGDB,
            help="Fetch missing information from PeeringDB if they are "
                 "not included in the Euro-IX JSON member list file.",
            dest="merge_from_peeringdb")

        parser.add_argument(
            "--merge-from-custom-file",
            help="Merge a custom set of client settings from "
                 "FILE into the clients generated from the "
                 "Euro-IX JSON file. Can be used to override "
                 "or enrich the clients which are automatically "
                 "built by this tool with some custom local "
                 "settings. For more info on how it works and "
                 "the syntax of the custom file, run this command "
                 "again with --help-merge-from-custom-file.",
            metavar="FILE",
            type=argparse.FileType('r'),
            dest="merge_from_custom_file")

        parser.add_argument(
            "--help-merge-from-custom-file",
            help="Prints additional info on how to use the "
                 "--merge-from-custom-file option.",
            action="store_true")

        parser.add_argument(
            "-o", "--output",
            type=argparse.FileType('w'),
            help="Output file. Default: stdout.",
            default=sys.stdout,
            dest="output_file")

    def run(self):
        if self.args.help_merge_from_custom_file:
            print(
                "\n"
                "When --merge-from-custom-file is used, a local "
                "YAML file is used to enrich the clients.yml file "
                "built using this command.\n"
                "This option can be used to handle local exceptions, "
                "to override settings or to add clients which are "
                "not included in the Euro-IX JSON used to built the "
                "original file.\n"
                "The format of the custom file must be the same of "
                "the regular clients.yml file; the 'ip' attribute "
                "of a client included inside the custom file is used "
                "to match the client on the original clients.yml "
                "build by this command. The content of the custom "
                "file is then merged into the definition of the "
                "original client.\n"
                "\n"
                "If a client is configured in the local custom file "
                "but is missing from the set of clients generated "
                "using the Euro-IX JSON file, then the local one is "
                "ignored, unless the 'add_if_missing' attribute is "
                "set to 'True': in that case, the custom client is "
                "added to the resulting clients.yml file.\n"
                "\n"
                "Example of a local custom file:\n"
                "\n"
                "clients:\n"
                "  - ip: 192.0.2.1\n"
                "    password: ""bgp_secret""\n"
                "  - ip: 192.0.2.2\n"
                "    cfg:\n"
                "      filtering:\n"
                "        irrdb:\n"
                "          as_sets:\n"
                "            - AS-TWO\n"
                "  - ip: 192.0.2.3\n"
                "    add_if_missing: True\n"
                "    asn: 3333\n"
            )
            return True

        euro_ix = EuroIXMemberList(self.args.url or self.args.input_file,
                                   program_config.get_dir("cache_dir"),
                                   program_config.get("cache_expiry"))

        if self.args.ixp_id:
            clients = euro_ix.get_clients(
                self.args.ixp_id, vlan_id=self.args.vlan_id,
                routeserver_only=self.args.routeserver_only,
                guess_custom_bgp_communities=self.args.guess_custom_bgp_communities,
                merge_from_peeringdb=self.args.merge_from_peeringdb)
            res = {"clients": clients}

            if self.args.merge_from_custom_file:
                try:
                    merge_clients(res, self.args.merge_from_custom_file)
                except ARouteServerError as e:
                    raise ARouteServerError(
                        "An error occurred while processing the custom "
                        "clients file provided via "
                        "--merge-from-custom-file: {}".format(e)
                    )

            comments = []
            comments.append("# Data fetched from {} at {} UTC".format(
                self.args.url or self.args.input_file.name,
                datetime.datetime.utcnow().isoformat()
            ))
            comments.append("# IXP ID: {}".format(self.args.ixp_id))
            if self.args.vlan_id:
                comments.append("# VLAN ID: {}".format(self.args.vlan_id))
            self.args.output_file.write("\n".join(comments) + "\n")
            yaml.safe_dump(res, self.args.output_file, default_flow_style=False)

            if self.args.guess_custom_bgp_communities and \
                euro_ix.unique_custom_communities:
                comments = []
                comments.append("# The following custom BGP communities must")
                comments.append("# be declared within the general.yml file:")
                comments.append("#  custom_communities:")
                for prefix in euro_ix.unique_custom_communities:
                    comms = euro_ix.unique_custom_communities[prefix]
                    if comms:
                        for comm in comms:
                            comments.append("#    {}:".format(comm))
                            comments.append("#      std:")
                            comments.append("#      ext:")
                            comments.append("#      lrg:")
                self.args.output_file.write("\n".join(comments) + "\n")
        else:
            euro_ix.print_infrastructure_list(self.args.output_file)

        return True



# Copyright (C) 2017-2023 Pier Carlo Chiodi
# Copyright (C) 2021 Ene Alin Gabriel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
from collections import OrderedDict
import datetime
import json
import logging
import os
import sys
import yaml

from .base import ARouteServerCommand
from ..config.asns import ConfigParserASNS
from ..config.clients import ConfigParserClients
from ..config.program import program_config
from ..errors import ARouteServerError, MissingFileError, IXFMemberListFromClientsMergeFileError

class IXFMemberListFromClientsCommand(ARouteServerCommand):

    COMMAND_NAME = "ixf-member-export"
    COMMAND_HELP = ("Build an IX-F Member Export JSON file "
                    "from the clients.")
    NEEDS_CONFIG = True

    @classmethod
    def add_arguments(cls, parser):
        super(IXFMemberListFromClientsCommand, cls).add_arguments(parser)

        parser.add_argument(
            "shortname",
            help="Short name of the IXP.")

        parser.add_argument(
            "ixf_id",
            type=int,
            help="IXP ID from the IX-F database.",
            metavar="IXF_ID")

        parser.add_argument(
            "--clients",
            help="Route server clients configuration file. "
                 "Default: the one set in the program configuration "
                 "file will be used.",
            metavar="FILE",
            dest="cfg_clients")

        parser.add_argument(
            "--ixp_id", "--ixp-id",
            type=int,
            default=0,
            help="The numeric identifier used by the IX to identify the "
                 "infrastructure for which the list of clients is generated. "
                 "Default: 0",
            dest="ixp_id")

        parser.add_argument(
            "--vlan-id",
            type=int,
            default=0,
            help="The VLAN ID used to set the connection_list.vlan_list "
                 "objects within the generated file. "
                 "Please note: this is not necessarily the 802.1q "
                 "tag but, generally speaking, it is the numeric identifier "
                 "used by the IX to identify the VLAN in its IX-F Member "
                 "Export JSON file.",
            dest="vlan_id")

        parser.add_argument(
            "--merge-file",
            type=argparse.FileType('r'),
            help="An optional JSON or YML file whose content will be merged "
                 "into the output generated by this command. This can be used "
                 "to add extra information to the JSON file generated by "
                 "ARouteServer, for example to include attributes that are "
                 "required on PeeringDB but that are not present in "
                 "ARouteServer (like IX country, full name, URL...). Check "
                 "https://arouteserver.readthedocs.io/en/latest/USAGE.html"
                 "#ixf-member-export-command for more details on how to use "
                 "it.",
            dest="merge_file")

        parser.add_argument(
            "-o", "--output",
            type=argparse.FileType('w'),
            help="Output file. Default: stdout.",
            default=sys.stdout,
            dest="output_file")

    @staticmethod
    def get_member_list(asns, clients, ixp_id, vlan_id):
        members = {}
        for client in clients:
            asn = str(client["asn"])
            if asn not in members:
                members[asn] = {"vlan_list": []}

            if client["description"]:
                members[asn]["descr"] = client["description"]

            vlan_obj = OrderedDict()
            vlan_obj["vlan_id"] = vlan_id

            ipv4_ipv6 = "ipv6" if ":" in client["ip"] else "ipv4"

            vlan_obj[ipv4_ipv6] = {"address": client["ip"]}

            max_pref = client["cfg"]["filtering"]["max_prefix"][
                "limit_{}".format(ipv4_ipv6)]

            if max_pref:
                vlan_obj[ipv4_ipv6]["max_prefix"] = max_pref

            as_macros = client["cfg"]["filtering"]["irrdb"]["as_sets"]
            if not as_macros:
                if "AS{}".format(asn) in asns:
                    as_macros = asns["AS{}".format(asn)]["as_sets"]

            if as_macros:
                if len(as_macros) > 1:
                    logging.warning(
                        "Client {ip} (AS{asn}) is configured with more than "
                        "one AS-SETs: since the destination format supports "
                        "only one item, only the first one ({as_set}) will "
                        "be exported.".format(
                            ip=client["ip"], asn=client["asn"],
                            as_set=as_macros[0]
                        )
                    )
                vlan_obj[ipv4_ipv6]["as_macro"] = as_macros[0]

            members[asn]["vlan_list"].append(vlan_obj)

        res = []
        for member_asn in members:
            member = {"asnum": int(member_asn)}

            if "descr" in members[member_asn]:
                member["name"] = members[member_asn]["descr"]

            connection_list_entry = OrderedDict()
            connection_list_entry["ixp_id"] = ixp_id
            connection_list_entry["vlan_list"] = members[member_asn]["vlan_list"]
            member["connection_list"] = [connection_list_entry]
            res.append(member)

        return res

    @staticmethod
    def load_config_from_path(path):
        clients = ConfigParserClients()
        asns = ConfigParserASNS()
        try:
            if not os.path.isfile(path):
                raise MissingFileError(path)

            clients.load(path)
            asns.load(path)
        except ARouteServerError as e:
            raise ARouteServerError(
                "One or more errors occurred while loading "
                "clients file{}".format(
                    ": {}".format(str(e)) if str(e) else ""
                )
            )

        return asns, clients

    @staticmethod
    def build_json(path, ixp_id, shortname, vlan_id, ixf_id):
        asns, clients = \
            IXFMemberListFromClientsCommand.load_config_from_path(path)

        res = OrderedDict()
        res["version"] = "1.0"
        res["timestamp"] = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')

        ixp_list_entry = OrderedDict()
        ixp_list_entry["ixp_id"] = ixp_id
        ixp_list_entry["ixf_id"] = ixf_id
        ixp_list_entry["shortname"] = shortname
        ixp_list_entry["vlan"] = [{"id": vlan_id}]
        res["ixp_list"] = [ixp_list_entry]

        res["member_list"] = IXFMemberListFromClientsCommand.get_member_list(
            asns, clients, ixp_id, vlan_id
        )
        return res

    @staticmethod
    def apply_merge_file(dic, merge_file, args):
        ixp_id_from_args = args["ixp_id"]
        ixf_id_from_args = args["ixf_id"]
        vlan_from_args = args["vlan"]

        # If the 'ixp_list' key is set in the merge file, let's assume that
        # it's better configured and more comprehensive than the one that
        # is generated by ARouteServer. So, get rid of the one present in
        # the dictionary built above, and use the one provided by the user.
        if "ixp_list" in merge_file:
            # Perform some sanity checks to be sure that the file that we're going
            # to produce is still valid. This is needed because we're using
            # content provided by the user, that may be partial or invalid.

            ixp_list = merge_file["ixp_list"]

            if not ixp_list:
                raise IXFMemberListFromClientsMergeFileError("ixp_list is empty", **args)
            if not isinstance(ixp_list, list):
                raise IXFMemberListFromClientsMergeFileError("ixp_list is not a list", **args)
            if not len(ixp_list) == 1:
                raise IXFMemberListFromClientsMergeFileError("ixp_list contains more than one dictionary", **args)

            ixp_list_entry = ixp_list[0]

            for key in ("ixf_id", "ixp_id", "shortname"):
                if key not in ixp_list_entry:
                    raise IXFMemberListFromClientsMergeFileError(
                        "the '{}' key cannot be found in the ixp_list dictionary".format(key), **args
                    )

                if not ixp_list_entry[key]:
                    raise IXFMemberListFromClientsMergeFileError(
                        "the value of the '{}' key in the ixp_list dictionary is empty or not set".format(key), **args
                    )

            for key in ("ixf_id", "ixp_id"):
                if not isinstance(ixp_list_entry[key], int):
                    raise IXFMemberListFromClientsMergeFileError(
                        "the value of the '{}' key in the ixp_list dictionary must be an integer".format(key), **args
                    )

            if ixp_list_entry["ixp_id"] != ixp_id_from_args:
                raise IXFMemberListFromClientsMergeFileError(
                    "the value of 'ixp_id' from the merge file ({ixp_id_from_merge_file}) doesn't match "
                    "the value from the command line arguments ({ixp_id_from_args}); please be sure that "
                    "both correspond, either adjusting the content of the merge file or passing the "
                    "--ixp_id {ixp_id_from_merge_file} option to the arouteserver ixf-member-export "
                    "command +".format(
                        ixp_id_from_merge_file=ixp_list_entry["ixp_id"],
                        ixp_id_from_args=ixp_id_from_args
                    ), **args
                )

            if ixp_list_entry["ixf_id"] != ixf_id_from_args:
                raise IXFMemberListFromClientsMergeFileError(
                    "the value of 'ixf_id' from the merge file ({ixf_id_from_merge_file}) doesn't match "
                    "the value from the command line arguments ({ixf_id_from_args}); please be sure that "
                    "both correspond, either adjusting the content of the merge file or passing the "
                    "{ixf_id_from_merge_file} value to the arouteserver ixf-member-export "
                    "command +".format(
                        ixf_id_from_merge_file=ixp_list_entry["ixf_id"],
                        ixf_id_from_args=ixf_id_from_args
                    ), **args
                )

            if "vlan" in ixp_list_entry:
                if not isinstance(ixp_list_entry["vlan"], list):
                    raise IXFMemberListFromClientsMergeFileError(
                        "the 'vlan' key is set, but it's not a list of dictionaries", **args
                    )

                if not all(isinstance(_, dict) for _ in ixp_list_entry["vlan"]):
                    raise IXFMemberListFromClientsMergeFileError(
                        "the 'vlan' key is set, but it's not a list of dictionaries", **args
                    )

                vlan_ids_from_merge_file = set([
                    _["id"] for _ in ixp_list_entry["vlan"] if "id" in _
                ])

                if args["vlan"] not in vlan_ids_from_merge_file:
                    raise IXFMemberListFromClientsMergeFileError(
                        "the VLAN ID {vlan_from_args} that was passed via the command line option --vlan-id "
                        "is not present in the list of VLANs from the merge file; please be sure that "
                        "the VLAN ID that is supplied as the command argument is present inside the "
                        "'ixp_list.vlan' key of the merge file (VLAN IDs found in the merge file: "
                        "{vlan_ids_from_merge_file})".format(
                            vlan_from_args=vlan_from_args,
                            vlan_ids_from_merge_file=", ".join(map(str, vlan_ids_from_merge_file))
                        ), **args
                    )

            # Now, we use the 'ixp_list' provided via the merge-file to replace
            # the content that was generated by ARouteServer. We do that by
            # removing "our" ixp_list, so that the one from the merge file will
            # be used instead.
            del dic["ixp_list"]

        # Merge "our" dictionary into the merge-file, to be sure that the content
        # from the former will be preferred over the content provided by the user.
        return {**merge_file, **dic}

    def run(self):
        path = self.args.cfg_clients or program_config.get("cfg_clients")

        dic = self.build_json(path, self.args.ixp_id, self.args.shortname,
                              self.args.vlan_id, self.args.ixf_id)

        if "--ixp_id" in sys.argv:
            logging.warning(
                "Deprecation notice: the '--ixp_id' option is now deprecated in favor "
                "of '--ixp-id' (dash instead of underscore). Please adjust your tooling "
                "to use the new version of the option. In the future, the '--ixp_id' "
                "will be removed."
            )

        if self.args.merge_file:
            merge_file_raw_content = self.args.merge_file.read()
            try:
                merge_file = yaml.safe_load(merge_file_raw_content)
            except Exception as e:
                yaml_exception = str(e)
                try:
                    merge_file = json.loads(merge_file_raw_content)
                except Exception as e:
                    json_exception = str(e)

                    raise ARouteServerError(
                        "The --merge-file is not in a valid YML or JSON format.\n\n"
                        "YAML error: {}\n\nJSON error: {}".format(
                            yaml_exception, json_exception
                        )
                    )

            args = {
                "ixp_id": self.args.ixp_id,
                "ixf_id": self.args.ixf_id,
                "shortname": self.args.shortname,
                "vlan": self.args.vlan_id
            }

            output_dic = self.apply_merge_file(dic, merge_file, args)
        else:
            output_dic = dic

        json.dump(output_dic, self.args.output_file, indent=2)
        return True



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import ARouteServerCommand

from ..config.program import program_config

class SetupCommand(ARouteServerCommand):

    COMMAND_NAME = "setup"
    COMMAND_HELP = ("Perform the setup of the system by copying "
                    "configuration files and templates in the proper "
                    "directories. Confirmation before each action will "
                    "be asked.")

    @classmethod
    def add_arguments(cls, parser):
        super(SetupCommand, cls).add_arguments(parser)

        parser.add_argument(
            "--dest-dir",
            type=str,
            help="Directory where the program's configuration files and "
                 "templates will be stored.",
            dest="dest_dir")

    def run(self):
        return program_config.setup(destination_directory=self.args.dest_dir)



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import time

from .kvm import KVMInstance
from .docker import DockerInstance
from .instances import Route, BGPSpeakerInstance, InstanceNotRunning

class OpenBGPDRoute(Route):

    def _parse_bgp_communities(self, communities):
        if not communities:
            return []

        res = []
        for bgp_comm in communities.split(" "):
            if bgp_comm == "BLACKHOLE":
                res.append("65535:666")
                continue
            if bgp_comm == "GRACEFUL_SHUTDOWN":
                res.append("65535:0")
                continue
            res.append(bgp_comm)
        return res

    def _parse_ext_bgp_communities(self, communities):
        if not communities:
            return []

        res = []

        # If ', ' is found in the value of 'communities', it
        # means we're processing a pre 6.6 output format.
        if ", " in communities:
            for bgp_comm in communities.split(", "):
                parts = bgp_comm.split(" ")
                res.append("{}:{}".format(parts[0], parts[1]))

            return res

        # If ', ' is not found, it's either a post 6.6 format
        # where different communities are separated just by a
        # space, or it's just one single community:
        # soo 65535:65281
        # rt 64537:10
        # rt 64537:10 rt 64538:20

        next_field_should_be_comm_type = True
        comm_type = ""
        comm = ""

        for part in communities.split():
            if next_field_should_be_comm_type:
                if part not in ("rt", "soo", "ovs"):
                    raise ValueError(
                        "Error while processing the extended communities "
                        "string '{}': expected 'rt', 'soo' or 'ovs', but '{}' "
                        "found.".format(communities, part)
                    )
                comm_type = part
                next_field_should_be_comm_type = False
            else:
                if comm_type != "ovs" and ":" not in part:
                    raise ValueError(
                        "Error while processing the extended communities "
                        "string '{}': ':' not found in the community '{}' "
                        "part.".format(communities, part)
                    )
                comm = part

                res.append("{}:{}".format(comm_type, comm))

                next_field_should_be_comm_type = True

        return res

class OpenBGPDInstance(object):
    """This class implements an interface to OpenBGPD client.

    Only methods needed to interact with the OpenBGPD bgpctl
    client are implemented here. The methods which interact with
    the underlying OS are offloaded to the other classes.

    This class is supposed to be used in conjunction with
    KVMInstance or DockerInstance since it uses some of
    their methods and properties.
    """

    MESSAGE_LOGGING_SUPPORT = False

    def __init__(self):
        self.neighbors_status = None
        self.routes = {}

    def _get_neighbors_status(self, force_update=False):
        if force_update:
            self.neighbors_status = None

        if self.neighbors_status:
            return

        self.neighbors_status = []

        out = self.run_cmd("bgpctl -n show summary terse")
        lines = out.split("\n")

        for line in lines:
            if not line.strip():
                continue

            parts = line.split(" ")
            if len(parts) != 3:
                raise Exception(
                    "{}: can't parse neighbor status: '{}'".format(
                        self.name, line
                    )
                )

            ip = parts[0].strip()
            asn = int(parts[1].strip())
            status = parts[2].strip()
            self.neighbors_status.append({
                "ip": ip,
                "is_up": "Established" in status,
                "asn": asn,
            })

        if not self.neighbors_status:
            raise Exception(
                "{}: can't get neighbors status".format(self.name)
            )

    def get_bgp_session(self, other_inst_or_ip, force_update=False):
        if isinstance(other_inst_or_ip, BGPSpeakerInstance):
            other_inst_ip = other_inst_or_ip.ip
        else:
            other_inst_ip = other_inst_or_ip
        self._get_neighbors_status(force_update=force_update)
        for neighbor in self.neighbors_status:
            if neighbor["ip"] == other_inst_ip:
                return neighbor
        return None

    def _parse_routes(self, out):
        regex = r"BGP routing table entry for ([0-9\.\:a-f]+/[0-9]+)"

        route = {}
        routes = []

        lines = out.split("\n")
        last_line_new_route = False
        for line_raw in lines:
            line = line_raw.strip()
            if not line:
                continue

            match = re.search(regex, line, re.MULTILINE)

            if match:
                if route:
                    routes.append(OpenBGPDRoute(**route))
                    route = {}
                route["prefix"] = match.group(1)
                last_line_new_route = True
                continue

            if last_line_new_route:
                raw_as_path = line
                if "{" in raw_as_path:
                    # Stripping as_set in strings like this:
                    #   222 333 { 333 333 }
                    as_path = raw_as_path[0:raw_as_path.index("{")].strip()
                    as_set = raw_as_path[raw_as_path.index("{") + 1:-1].strip()
                else:
                    as_path = raw_as_path
                    as_set = None
                asns = as_path.split(" ")
                for asn in asns:
                    if not asn.isdigit():
                        raise Exception(
                            "Error parsing {}: invalid AS_PATH: {}".format(
                                route["prefix"], line
                            )
                        )
                route["as_path"] = as_path
                route["as_set"] = as_set
                last_line_new_route = False
                continue

            if line.startswith("Nexthop"):
                parts = line.split(" ")
                route["next_hop"] = parts[1]
                route["via"] = parts[5]
            elif line.startswith("Origin"):
                route["best"] = "best" in line

                match = re.search("localpref ([0-9]+)", line)
                route["localpref"] = int(match.group(1))
            elif line.startswith("Communities:"):
                route["std_comms"] = line.split(": ")[1]
            elif line.lower().startswith("ext. communities:"):
                route["ext_comms"] = line.split(": ")[1]
            elif line.startswith("Large Communities:"):
                route["lrg_comms"] = line.split(": ")[1]
            elif line.startswith("OTC:"):
                route["otc"] = line.split(": ")[1]
            last_line_new_route = False

        if last_line_new_route:
            raise Exception("Error parsing routes: last route {}".format(
                route["prefix"] if route else "unknown"))

        if route:
            routes.append(OpenBGPDRoute(**route))
            route = {}

        return routes

    def _get_routes_from_neighbor(self, ip):
        out = self.run_cmd(
            "bgpctl -n show rib in neighbor {} detail".format(ip)
        )
        return self._parse_routes(out)

    def _get_routes_from_main(self):
        out = self.run_cmd(
            "bgpctl -n show ip bgp detail"
        )
        routes = self._parse_routes(out)

        # Add rejected routes.
        out = self.run_cmd(
            "bgpctl -n show ip bgp detail error"
        )
        for rejected_route in self._parse_routes(out):
            rejected_route.filtered = True
            routes.append(rejected_route)

        return routes

    def _get_routes_from_all_sources(self):
        self.routes = {
            "main": [],
        }
        self.routes["main"] = self._get_routes_from_main()

    def _route_is_in(self, route, route_list):
        for other_route in route_list:
            if route.prefix == other_route.prefix and \
                route.via == other_route.via and \
                route.as_path == other_route.as_path and \
                route.next_hop == other_route.next_hop and \
                route.std_comms == other_route.std_comms and \
                route.ext_comms == other_route.ext_comms and \
                route.lrg_comms == other_route.lrg_comms:
                return True
        return False

    def clear_cached_routes(self):
        self.routes = {}

    def get_routes(self, prefix, include_filtered=False, only_best=False):
        if include_filtered and only_best:
            raise Exception("Can't set both include_filtered and only_best")

        if not self.routes:
            self._get_neighbors_status(force_update=True)
            self._get_routes_from_all_sources()

        res = []

        for route in self.routes["main"]:
            if prefix and route.prefix.lower() != prefix.lower():
                continue
            if only_best and not route.best:
                continue
            if not self._route_is_in(route, res):
                res.append(route)

        return res


    def log_contains(self, s):
        return True

    def log_contains_errors(self, allowed_errors=[], list_errors=False):
        if list_errors:
            return False, ""
        return False

class OpenBGPDClassicInstance(OpenBGPDInstance, KVMInstance):
    """This class implements OpenBGPD-specific methods.

    This class is derived from :class:`KVMInstance`, that implements
    some kvm-specific methods to start/stop the instance and to run
    commands on it.

    The VIRSH_DOMAINNAME attribute must be set by derived classes on the
    basis of the specific version of OpenBSD they represent.
    """

    VIRSH_DOMAINNAME = None

    def __init__(self, *args, **kwargs):
        OpenBGPDInstance.__init__(self)
        KVMInstance.__init__(self, *args, **kwargs)

    def _graceful_shutdown(self):
        self.run_cmd("shutdown -h -p now")
        return True

    def restart(self):
        """Restart OpenBGPD.

        Updates the configuration files, then executes '/etc/rc.d/bgpd stop'
        and then '/etc/rc.d/bgpd -f start'.
        """
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        try:
            self.run_cmd("mkdir /etc/bgpd")
        except:
            pass

        self._mount_files()

        self.run_cmd("chmod 0600 /etc/bgpd.conf")
        self.run_cmd("touch /etc/bgpd/placeholder")
        self.run_cmd("chmod 0600 /etc/bgpd/*")

        self.run_cmd("/etc/rc.d/bgpd stop")
        time.sleep(5)
        self.run_cmd("ndp -c | true")
        self.run_cmd("bgpd -dn")
        self.run_cmd("/etc/rc.d/bgpd -f start")
        time.sleep(5)

        return True

    def reload_config(self):
        """Reload OpenBGPD configuration.

        Updates the configuration files, then executes '/etc/rc.d/bgpd reload'.
        """
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        self._mount_files()

        self.run_cmd("bgpd -dn")
        self.run_cmd("/etc/rc.d/bgpd reload")
        self.run_cmd("ndp -c | true")
        time.sleep(5)

        return True

    def log_contains(self, s):
        return True

    def log_contains_errors(self, allowed_errors=[], list_errors=False):
        if list_errors:
            return False, ""
        return False

class OpenBGPDPortableInstance(OpenBGPDInstance, DockerInstance, BGPSpeakerInstance):
    """This class implements OpenBGPD-specific methods for the Portable edition.

    This class is derived from :class:`DockerInstance`, that implements
    some docker-specific methods to start/stop the instance and to run
    commands on it.

    The VIRSH_DOMAINNAME attribute must be set by derived classes on the
    basis of the specific version of OpenBSD they represent.
    """

    def __init__(self, *args, **kwargs):
        OpenBGPDInstance.__init__(self)
        DockerInstance.__init__(self, *args, **kwargs)

    def restart(self):
        """Restart OpenBGPD.

        Updates the configuration file, then clear all the neighbors.
        """

        self.reload_config()

        self._get_neighbors_status(force_update=True)
        for neighbor in self.neighbors_status:
            ip = neighbor["ip"]
            self.run_cmd("bgpctl neighbor {} clear".format(ip))

        time.sleep(5)

        return True

    def reload_config(self):
        """Reload OpenBGPD configuration.

        Executes 'bgpctl reload'.
        """
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        try:
            res = self.run_cmd("bgpctl reload")
        except:
            res = ""

        if "request processed" in res:
            time.sleep(5)
            return True

        return False

    def _get_start_cmd(self):
        return "bgpd -f /etc/bgpd.conf -d"

class OpenBGPD60Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd60"

    TAG = "openbgpd60"

class OpenBGPD61Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd61"

    TAG = "openbgpd61"

class OpenBGPD62Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd62"

    TAG = "openbgpd62"

class OpenBGPD63Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd63"

    TAG = "openbgpd63"

class OpenBGPD64Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd64"

    TAG = "openbgpd64"

class OpenBGPD65Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd65"

    TAG = "openbgpd65"

    BGP_SPEAKER_VERSION = "6.5"
    TARGET_VERSION = "6.5"

class OpenBGPD66Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd66"

    TAG = "openbgpd66"

    BGP_SPEAKER_VERSION = "6.6"
    TARGET_VERSION = "6.6"

class OpenBGPD67Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd67"

    TAG = "openbgpd67"

    BGP_SPEAKER_VERSION = "6.7"
    TARGET_VERSION = "6.7"

class OpenBGPD68Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd68"

    TAG = "openbgpd68"

    BGP_SPEAKER_VERSION = "6.8"
    TARGET_VERSION = "6.8"

class OpenBGPD69Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd69"

    TAG = "openbgpd69"

    BGP_SPEAKER_VERSION = "6.9"
    TARGET_VERSION = "6.9"

class OpenBGPD70Instance(OpenBGPDClassicInstance):

    VIRSH_DOMAINNAME = "arouteserver_openbgpd70"

    TAG = "openbgpd70"

    BGP_SPEAKER_VERSION = "7.0"
    TARGET_VERSION = "7.0"

OpenBGPDPreviousInstance = OpenBGPD68Instance
OpenBGPDLatestInstance = OpenBGPD70Instance

class OpenBGPD65PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:6.5p1"

    TAG = "openbgpd65p"

class OpenBGPD66PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:6.6p0"

    TAG = "openbgpd66p"

    BGP_SPEAKER_VERSION = "6.6p0"
    TARGET_VERSION = "6.6"

class OpenBGPD67PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:6.7p0"

    TAG = "openbgpd67p"

    BGP_SPEAKER_VERSION = "6.7p0"
    TARGET_VERSION = "6.7"

class OpenBGPD68PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:6.8p1"

    TAG = "openbgpd68p"

    BGP_SPEAKER_VERSION = "6.8p1"
    TARGET_VERSION = "6.8"

class OpenBGPD69PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:6.9p0-patches"

    TAG = "openbgpd69p"

    BGP_SPEAKER_VERSION = "6.9p0"
    TARGET_VERSION = "6.9"

class OpenBGPD70PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.0p0"

    TAG = "openbgpd70p"

    BGP_SPEAKER_VERSION = "7.0p0"
    TARGET_VERSION = "7.0"

class OpenBGPD71PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.1p0"

    TAG = "openbgpd71p"

    BGP_SPEAKER_VERSION = "7.1p0"
    TARGET_VERSION = "7.1"


class OpenBGPD72PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.2"

    TAG = "openbgpd72p"

    BGP_SPEAKER_VERSION = "7.2"
    TARGET_VERSION = "7.2"


class OpenBGPD73PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.3"

    TAG = "openbgpd73p"

    BGP_SPEAKER_VERSION = "7.3"
    TARGET_VERSION = "7.3"


class OpenBGPD74PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.4"

    TAG = "openbgpd74p"

    BGP_SPEAKER_VERSION = "7.4"
    TARGET_VERSION = "7.4"


class OpenBGPD75PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.5"

    TAG = "openbgpd75p"

    BGP_SPEAKER_VERSION = "7.5"
    TARGET_VERSION = "7.5"


class OpenBGPD76PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.6"

    TAG = "openbgpd76p"

    BGP_SPEAKER_VERSION = "7.6"
    TARGET_VERSION = "7.6"


class OpenBGPD77PortableInstance(OpenBGPDPortableInstance):

    DOCKER_IMAGE = "pierky/openbgpd:7.7"

    TAG = "openbgpd77p"

    BGP_SPEAKER_VERSION = "7.7"
    TARGET_VERSION = "7.7"


OpenBGPDPortablePreviousInstance = OpenBGPD76PortableInstance
OpenBGPDPortableLatestInstance = OpenBGPD77PortableInstance



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap

from .base import ARouteServerCommand

from ..config.program import program_config

class VerifyTemplatesCommand(ARouteServerCommand):

    COMMAND_NAME = "verify-templates"
    COMMAND_HELP = ("Verify if templates are aligned with the current "
                    "version of the program and print debugging info.")
    NEEDS_CONFIG = True

    def run(self):
        errors = program_config.verify_templates()
        all_right = errors == []
        if not all_right:
            print("\n".join(textwrap.wrap(
                  "One or more templates are not aligned with those "
                  "distributed with the current version of the program:\n")))
            print("")
            for err in errors:
                print(" - "
                      "\n   ".join(textwrap.wrap(err)))
            print("")
            print("\n".join(textwrap.wrap(
                  "Unexpected behaviours may occur during the configuration "
                  "building process: if you experience issues please consider "
                  "running the 'arouteserver setup-templates' command to "
                  "restore the original templates that are distributed with "
                  "the current version of the program.")))
            print("\n".join(textwrap.wrap(
                  "To customize the configuration of the route server with "
                  "your own options, please consider using 'Site-specific "
                  "custom configuration files' instead of editing the "
                  "template files: more info at "
                  "https://arouteserver.readthedocs.io/en/latest/CONFIG.html"
                  "#site-specific-custom-config")))
        else:
            print("Everything is fine; templates are aligned with those "
                  "expected by the current version of the program.")
        return all_right



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
import logging
import yaml

from .base import ConfigParserBase, convert_deprecated
from .validators import *
from ..errors import ConfigError, ARouteServerError


class ConfigParserClients(ConfigParserBase):

    ROOT = "clients"

    def __init__(self, general_cfg=None):
        ConfigParserBase.__init__(self)
        self.general_cfg = general_cfg
        self.asn3216_map = {}

    def parse(self):

        def get_client_descr(client):
            client_descr = ""
            if client.get("asn", None):
                client_descr += "AS{}".format(client["asn"])
            if client.get("ip", None):
                client_descr += " " + client["ip"]
            if not client_descr:
                client_descr = "unknown client"
            return client_descr

        if "clients" not in self.cfg:
            raise ConfigError("Missing top 'clients' statement.")
        if "asns" in self.cfg:
            del self.cfg["asns"]

        errors = False

        schema = {
            "asn": ValidatorASN(),
            "ip": ValidatorIPAddr(),
            "description": ValidatorText(mandatory=False),
            "password": ValidatorText(mandatory=False),
            "16bit_mapped_asn": ValidatorPrivateASN16bit(mandatory=False),
            "cfg": {
                "prepend_rs_as": ValidatorBool(mandatory=False),
                "passive": ValidatorBool(mandatory=False),
                "gtsm": ValidatorBool(mandatory=False),
                "multihop": ValidatorUInt(mandatory=False),
                "add_path": ValidatorBool(mandatory=False),
                "filtering": {
                    "next_hop": {
                        "policy": ValidatorOption("policy",
                                                  ("strict", "same-as",
                                                   "authorized_addresses"),
                                                  mandatory=False),
                        "authorized_addresses_list": ValidatorListOf(
                            ValidatorIPAddr, mandatory=False
                        )
                    },
                    "ipv4_pref_len": ValidatorIPMinMaxLen(4, mandatory=False),
                    "ipv6_pref_len": ValidatorIPMinMaxLen(6, mandatory=False),
                    "max_as_path_len": ValidatorMaxASPathLen(mandatory=False),
                    "reject_invalid_as_in_as_path": ValidatorBool(mandatory=False),
                    "irrdb": {
                        "as_sets": ValidatorListOf(ValidatorASSet,
                                                   mandatory=False),
                        "enforce_origin_in_as_set": ValidatorBool(mandatory=False),
                        "enforce_prefix_in_as_set": ValidatorBool(mandatory=False),
                        "white_list_pref": ValidatorListOf(
                            ValidatorPrefixListEntry, mandatory=False,
                        ),
                        "white_list_asn": ValidatorASNList(mandatory=False),
                        "white_list_route": ValidatorListOf(
                            ValidatorWhiteListRouteEntry, mandatory=False
                        )
                    },
                    "rpki_bgp_origin_validation": {
                        "enabled": ValidatorBool(mandatory=False),
                        "reject_invalid": ValidatorBool(mandatory=False),
                    },
                    "black_list_pref": ValidatorListOf(
                        ValidatorPrefixListEntry, mandatory=False,
                    ),
                    "max_prefix": {
                        "action": ValidatorOption(
                            "action",
                            ("shutdown", "restart", "block", "warning"),
                            mandatory=False
                        ),
                        "restart_after": ValidatorUInt(mandatory=False),
                        "count_rejected_routes": ValidatorBool(mandatory=False),
                        "peering_db": {
                            "enabled": ValidatorBool(mandatory=False),
                            "increment": {
                                "absolute": ValidatorUInt(mandatory=False),
                                "relative": ValidatorUInt(mandatory=False)
                            }
                        },
                        "limit_ipv4": ValidatorUInt(mandatory=False),
                        "limit_ipv6": ValidatorUInt(mandatory=False),
                    },
                    "roles": {
                        "enabled": ValidatorBool(mandatory=False),
                        "strict_mode": ValidatorBool(mandatory=False),
                        "local_role": ValidatorRole(mandatory=True, default="rs"),
                    },
                    "reject_policy": {
                        "policy": ValidatorOption("reject_policy",
                                                  ("reject", "tag", "tag_and_reject"),
                                                  mandatory=False)
                    },
                },
                "blackhole_filtering" : {
                   "announce_to_client": ValidatorBool(mandatory=False),
                },
                "graceful_shutdown": {
                    "enabled": ValidatorBool(mandatory=False)
                },
                "attach_custom_communities": ValidatorListOf(ValidatorText,
                                                             mandatory=False),
                "custom_options": ValidatorClientCustomOptions(mandatory=False)
            }
        }

        # Split configurations with more than one IP address into
        # multiple clients
        for client in self.cfg["clients"]:
            if "ip" in client:
                if isinstance(client["ip"], list):
                    for ip in client["ip"]:
                        client_clone = deepcopy(client)
                        client_clone["ip"] = ip
                        self.cfg["clients"].append(client_clone)
                    client["to_be_removed"] = True
        self.cfg["clients"] = [c for c in self.cfg["clients"] if "to_be_removed" not in c]

        # Clients' config validation
        for client in self.cfg["clients"]:
            client_descr = get_client_descr(client)

            try:
                # Convert next_hop_policy (< v0.6.0) into the new format
                if "cfg" in client:
                    convert_deprecated(client["cfg"])

                ConfigParserBase.validate(schema, client, "clients")
            except ARouteServerError as e:
                err_msg = ("One or more errors occurred while processing "
                           "the client configuration for "
                           "'{}'".format(client_descr))
                if str(e):
                    err_msg += ": " + str(e)
                logging.error(err_msg)
                raise ConfigError()

        def inherit_from_general_cfg(dest, src, schema):
            for k in schema:
                if isinstance(schema[k], dict):
                    if k in src:
                        inherit_from_general_cfg(dest[k], src[k], schema[k])
                else:
                    if k not in dest or dest[k] is None:
                        if k in src:
                            dest[k] = src[k]

        # Inherit missing options from the general configuration.
        if self.general_cfg:
            for client in self.cfg["clients"]:
                inherit_from_general_cfg(client["cfg"], self.general_cfg, schema["cfg"])

        # Duplicate IP addresses?
        unique_ip = []
        for client in self.cfg["clients"]:
            ip = client["ip"]
            if ip in unique_ip:
                logging.error(
                    "Duplicate IP address found: {}.".format(ip)
                )
                errors = True
            else:
                unique_ip.append(ip)

        # Clients with...
        # - next_hop.policy == "authorized_addresses" AND
        # - no authorized IP addresses
        # ... or with...
        # - "authorized_addresses_list" AND
        # - next_hop.policy != "authorized_addresses"
        for client in self.cfg["clients"]:
            client_descr = get_client_descr(client)
            next_hop = client["cfg"]["filtering"]["next_hop"]

            if next_hop["policy"] == "authorized_addresses" and \
                not next_hop["authorized_addresses_list"]:

                logging.error("The next_hop policy for client {} "
                              "is set to 'authorized_addresses' but "
                              "the list of authorized IP addresses "
                              "('authorized_addresses_list') is empty".format(
                                  client_descr
                                ))
                errors = True

            if next_hop["policy"] != "authorized_addresses" and \
                "authorized_addresses_list" in next_hop and \
                next_hop["authorized_addresses_list"] is not None:

                logging.error("The next_hop policy for client {} "
                              "is not 'authorized_addresses' but "
                              "the 'authorized_addresses_list' option "
                              "is set".format(
                                  client_descr
                                ))
                errors = True

        # Custom BGP communities must be declared within the general cfg
        for client in self.cfg["clients"]:
            client_descr = get_client_descr(client)

            if not client["cfg"]["attach_custom_communities"]:
                continue

            for comm in client["cfg"]["attach_custom_communities"]:
                if self.general_cfg and \
                    comm not in self.general_cfg["custom_communities"]:

                    logging.error("The custom BGP community {} "
                                  "referenced on client {} is not declared on "
                                  "the general configuration.".format(
                                    comm, client_descr
                                    ))
                    errors = True

        # Build the asn3216_map and check 16bit_mapped_asn.
        for client in self.cfg["clients"]:
            client_descr = get_client_descr(client)

            client_asn = int(client["asn"])
            mapped_asn = client.get("16bit_mapped_asn")

            if not mapped_asn:
                continue

            if client_asn <= 65535 and mapped_asn:
                logging.error(
                    "The '16bit_mapped_asn' option can be set only for "
                    "32bit ASNs, but it was set for the 16bit ASN "
                    "client {}".format(client_descr)
                )
                errors = True
                continue

            if client_asn in self.asn3216_map:
                if self.asn3216_map[client_asn] != mapped_asn:
                    logging.error(
                        "The '16bit_mapped_asn' option of client {} was "
                        "set to {}, but previously another 16bit ASN was "
                        "already used for that ASN: {}".format(
                            client_descr,
                            mapped_asn,
                            self.asn3216_map[client_asn]
                        )
                    )
                    errors = True
                    continue
            else:
                for other_32bit_asn, other_16bit_asn in self.asn3216_map.items():
                    if other_16bit_asn == mapped_asn and other_32bit_asn != client_asn:
                        logging.error(
                            "The 16bit ASN {} was used for the "
                            "'16bit_mapped_asn' option of client {}, but "
                            "previously it was already used to map the 32bit "
                            "ASN {} of another client".format(
                                mapped_asn,
                                client_descr,
                                other_32bit_asn
                            )
                        )
                        errors = True
                        break
                else:
                    self.asn3216_map[client_asn] = mapped_asn

        if errors:
            raise ConfigError()

def merge_clients(original, custom_file):
    try:
        new = yaml.safe_load(custom_file)
    except Exception as e:
        raise ConfigError(
            "Error while loading the YAML file: {e}".format(e=e)
        )

    if not new:
        raise ConfigError("The YAML file is empty")

    if "clients" not in new:
        raise ConfigError(
            "The top level 'clients' key is missing from the "
            "the set of clients to be merged."
        )

    if not isinstance(new["clients"], list):
        raise ConfigError(
            "The top level 'clients' key must represent a list "
            "of clients."
        )

    for client_id, client in enumerate(new["clients"]):
        try:
            if "ip" not in client:
                raise ConfigError(
                    "'ip' not found"
                )

            if not isinstance(client["ip"], str):
                raise ConfigError(
                    "'ip' must be a string representing the IP "
                    "address of the original client to be updated."
                )


            # The value of the 'ip' key of the dictionary that will
            # be used below to find the corresponding client and
            # to update it is overriden here with the normalised
            # representation of that IP.
            # This is done in order to avoid mismatching of IPs
            # represented in different ways (lower/upper case,
            # exploded vs compact form), and also to avoid
            # changing the way the IP was represented in
            # the original client definition.
            client["ip"] = ValidatorIPAddr().validate(client["ip"])

            for original_client in original["clients"]:
                if original_client["ip"] != client["ip"]:
                    continue

                if "add_if_missing" in client:
                    raise ConfigError(
                        "client {ip} already exists in the original "
                        "list of clients, but it's also reported in "
                        "the set of clients to be merged with the "
                        "'add_if_missing' attribute set.".format(
                            ip=client["ip"]
                        )
                    )

                # A client having the same IP is present in the
                # list of original clients. It's the one to update
                # with the settings from the custom one.
                original_client.update(client)
                break
            else:
                # No clients to update were found.
                # If the custom one has the 'add_if_missing' key
                # set to True, the custom one can be added to the
                # list of clients.
                if client.pop("add_if_missing", False):
                    original["clients"].append(client)

        except ConfigError as e:
            raise ConfigError(
                "Error while processing the client n. {n} from "
                "the set of clients to be merged: {e}".format(
                    n=client_id + 1,
                    e=e
                )
            )
    try:
        new_res_after_merge = deepcopy(original)
        ConfigParserClients().load_from_dict(new_res_after_merge)
    except ARouteServerError:
        raise ConfigError(
            "Validation of the final clients file failed: "
            "check the logs for more details."
        )



# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import json
import logging
import re
import subprocess

from .cached_objects import CachedObject
from .config.validators import ValidatorPrefixListEntry
from .errors import IRRDBToolsError
from .ipaddresses import IPNetwork


TIMEDOUT_IRR_HOSTS = set()


class AS_SET_Bundle(object):

    @staticmethod
    def get_source(object_names):
        """Returns (source_to_use, is_same_for_all)

        In case of multiple objects, the first specific source found
        will be used. Otherwise source_to_use will be None (use the
        default sources).
        """

        sources = []
        for name in object_names:
            source_macro = name.split("::")
            if len(source_macro) > 1:
                sources.append(source_macro[0])
            else:
                sources.append(None)

        same_for_all = len(set(sources)) == 1

        not_none = [_ for _ in sources if _ is not None]
        if not_none:
            used_source = not_none[0].upper()
        else:
            used_source = None

        return (used_source, same_for_all)

    def __init__(self, object_names):
        assert isinstance(object_names, list)

        self.source, _ = self.get_source(object_names)

        self.object_names = sorted([n.upper() for n in set(object_names)])

        # id, internal unique identifier.
        buf = "_".join(self.object_names)
        hasher = hashlib.sha512()
        hasher.update(buf.encode("utf-8"))
        self.id = hasher.hexdigest()

        # descr, textual description of the bundle.
        # Do not use it in templates unless within comments.
        self.descr = ", ".join(self.object_names[:3])
        if len(self.object_names) > 3:
            self.descr += " and {} more".format(len(self.object_names) - 3)

        # name, brief textual representation of the bundle.
        # Only [a-zA-Z0-9_] characters.
        # Can be used in templates.
        if len(self.object_names) == 1:
            self.name = self.object_names[0]
        elif len(self.object_names) <= 3:
            self.name = "_".join(self.object_names)
        else:
            self.name = "{name}_and_{more}_more_{short_hash}".format(
                name=self.object_names[0],
                more=len(self.object_names) - 1,
                short_hash=self.id[:7]
            )
        self.name = re.sub("[^a-zA-Z0-9_]", "_", self.name)

        # The name will be used to generate config statements
        # like AS_SET_<name>_asns and AS_SET_<name>_prefixes.
        #
        # In BIRD there is a limit of 64 characters for symbols:
        #
        #   #define SYM_MAX_LEN 64
        #
        # so every name that, once concatenated to the prefix
        # and suffix above, will be longer than 64 characters
        # is here truncated, and a "random" tag is attached
        # to it in order to make it unique.
        #
        # 16 is the length of "AS_SET_" + "_prefixes", the longest
        # combination of prefix and suffix.

        if len(self.name) + 16 > 64:
            TAG_LEN = 5
            hash_str = hashlib.md5(self.name.encode("utf-8")).hexdigest()
            tag = hash_str[:TAG_LEN]
            max_name_len = 64 - 16 - TAG_LEN - 1
            self.name = self.name[:max_name_len] + "_" + tag

class IRRDBInfo(CachedObject, AS_SET_Bundle):

    BGPQ3_DEFAULT_HOST = ["rr.ntt.net", "rr1.ntt.net"]
    BGPQ3_DEFAULT_SOURCES = ("RIPE,APNIC,AFRINIC,ARIN,NTTCOM,ALTDB,"
                             "BBOI,BELL,JPIRR,LEVEL3,RADB,TC")
    BGPQ3_DEFAULT_TIMEOUT = 120
    EXPIRY_TIME_TAG = "irr_as_sets"

    def __init__(self, object_names, *args, **kwargs):
        assert isinstance(object_names, list)

        CachedObject.__init__(self, *args, **kwargs)
        self.bgpq3_path = kwargs.get("bgpq3_path")
        self.bgpq3_host = kwargs.get("bgpq3_host", self.BGPQ3_DEFAULT_HOST)
        self.bgpq3_sources = kwargs.get("bgpq3_sources",
                                        self.BGPQ3_DEFAULT_SOURCES)
        self.bgpq3_timeout = kwargs.get("bgpq3_timeout", self.BGPQ3_DEFAULT_TIMEOUT)

        self.bgpq = "bgpq4" if "bgpq4" in self.bgpq3_path else "bgpq3"

        AS_SET_Bundle.__init__(self, object_names)

    def _get_bgpq3_sources(self):
        if self.source:
            return "{},{}".format(self.source, self.bgpq3_sources)
        return self.bgpq3_sources

    def _get_bgpq3_names(self):
        res = []
        for name in self.object_names:
            source_macro = name.split("::")
            if len(source_macro) > 1:
                res.append(source_macro[1])
            else:
                res.append(source_macro[0])
        return res

    def _run_cmd(self, cmd):
        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

        try:
            out, err = proc.communicate(timeout=self.bgpq3_timeout)
        except subprocess.TimeoutExpired:
            proc.kill()
            proc.communicate()
            raise

        if proc.returncode != 0:
            err_msg = "{} exit code is {}".format(self.bgpq, proc.returncode)
            if err is not None and err.strip():
                err_msg += ", stderr: {}".format(err)
            raise ValueError(err_msg)

        if err is not None and err.strip():
            logging.warning("{} succeeded but an error was "
                            "printed when executing '{}': {}".format(
                                self.bgpq,
                                " ".join(cmd), err.strip()
                            ))

        return out

    def _run_query(self, args):
        hosts_to_use = [
            host
            for host in self.bgpq3_host
            if host not in TIMEDOUT_IRR_HOSTS
        ]

        if not hosts_to_use:
            raise IRRDBToolsError(
                "All the IRRD hosts timed out so far; there are no more hosts "
                "to use to perform the IRR queries."
            )

        num_of_hosts = len(hosts_to_use)

        attempt_n = 0
        for host in hosts_to_use:
            attempt_n += 1

            cmd = [self.bgpq3_path]
            cmd += ["-h", host]
            cmd += ["-S", self._get_bgpq3_sources()]
            if "bgpq4" not in self.bgpq3_path:
                cmd += ["-3"]
            cmd += ["-j"]
            cmd += args

            try:
                out = self._run_cmd(cmd)

                return json.loads(out.decode("utf-8"))
            except subprocess.TimeoutExpired:
                err_msg = (
                    "{} timed out while running the following command: '{}' "
                    "The host {} will not be used for the next IRR queries. "
                    "The timeout is {} seconds; to modify it, please "
                    "edit the program's configuration file (usually "
                    "arouteserver.yml) and change the 'bgpq3_timeout' setting."
                ).format(
                    self.bgpq,
                    " ".join(cmd),
                    host,
                    self.bgpq3_timeout
                )
                TIMEDOUT_IRR_HOSTS.add(host)
            except Exception as e:
                err_msg = (
                    "Error while parsing {} output "
                    "for the following command: '{}': {}".format(
                        self.bgpq,
                        " ".join(cmd), str(e)
                    )
                )

            if attempt_n == num_of_hosts:
                raise IRRDBToolsError(
                    "{} - No more attempts will be performed, all the "
                    "hosts in the list failed.".format(err_msg)
                )
            else:
                logging.warning(
                    "{} - Another attempt will be performed using the next "
                    "host in the list.".format(err_msg)
                )


class ASSet(IRRDBInfo):

    def load_data(self):
        logging.debug("Getting ASNs for "
                      "{} from IRRdb".format(self.descr))

        IRRDBInfo.load_data(self)

        # list of int
        self.asns = self.raw_data

    def _get_object_filename(self):
        return "{}-as_set.json".format(self.name)

    def _get_data(self):
        object_names = self._get_bgpq3_names()

        # If the list of objects to expand is made up by
        # an 'ASxxx' element only, avoid to run bgpq3 and
        # return only that ASN.
        if len(object_names) == 1 and \
            re.match("^AS[0-9]+$", object_names[0]):
            return [int(object_names[0][2:])]

        query_args = []
        query_args += ["-f", "1"]
        query_args += ["-l", "asn_list"]
        query_args += object_names

        try:
            data = self._run_query(query_args)
        except Exception as e:
            raise IRRDBToolsError(
                "Can't get list of authorized ASNs for {}: {}".format(
                    self.descr, str(e)
                )
            )

        return data["asn_list"]

class RSet(IRRDBInfo):

    def __init__(self, object_names, ip_ver, allow_longer_prefixes, **kwargs):
        IRRDBInfo.__init__(self, object_names, **kwargs)

        assert ip_ver in (4, 6)
        self.ip_ver = ip_ver
        self.allow_longer_prefixes = allow_longer_prefixes

    def load_data(self):
        logging.debug("Getting prefixes for {} IPv{} "
                      "from IRRdb".format(self.descr, self.ip_ver))

        IRRDBInfo.load_data(self)

        # list of dict as returned by ValidatorPrefixListEntry
        self.prefixes = self.raw_data

    def _get_object_filename(self):
        return "{}-r_set-ipv{}{}.json".format(
            self.name, self.ip_ver,
            "_and_more_specific" if self.allow_longer_prefixes else ""
        )

    def _get_data(self):
        query_args = []
        query_args += ["-4"] if self.ip_ver == 4 else ["-6"]
        query_args += ["-A"]
        query_args += ["-l", "prefix_list"]
        if self.allow_longer_prefixes:
            query_args += ["-R"]
            query_args += ["32"] if self.ip_ver == 4 else ["128"]
        query_args += self._get_bgpq3_names()

        try:
            data = self._run_query(query_args)
        except Exception as e:
            raise IRRDBToolsError(
                "Can't get authorized prefix list for {} IPv{}: {}".format(
                    self.descr, self.ip_ver, str(e)
                )
            )

        return [self._parse_prefix(prefix) for prefix in data["prefix_list"]]

    def _parse_prefix(self, raw):
        prefix = IPNetwork(raw["prefix"])
        res = {
            "prefix": prefix.ip,
            "length": prefix.prefixlen,
            "exact": raw["exact"] if "exact" in raw else False
        }
        if res["exact"]:
            res["ge"] = None
            res["le"] = None
        else:
            if "greater-equal" in raw:
                res["ge"] = raw["greater-equal"]
            else:
                res["ge"] = None

            if "less-equal" in raw:
                res["le"] = raw["less-equal"]
            else:
                res["le"] = None

        return ValidatorPrefixListEntry().validate(res)
